"""
Experimental Cortex-M7 microarchitecture model for SLOTHY

.. warning::

    The data in this module is approximate and may contain errors.
"""

# ################################ NOTE ########################################## #
#                                                                                  #
# WARNING: The data in this module is approximate and may contain errors.          #
#          They are _NOT_ an official software optimization guide for Cortex-M7.   #
#                                                                                  #
# Sources used in constructing this model:                                         #
# - ARMv7-M Architecture Reference Manual (ARM DDI 0403E.e)                        #
# - https://github.com/jnk0le/random/tree/master/pipeline%20cycle%20test#cortex-m7 #
# - https://www.quinapalus.com/cm7cycles.html                                      #
####################################################################################

from enum import Enum
from itertools import product
from slothy.helper import lookup_multidict
from slothy.targets.arm_v7m.arch_v7m import (
    find_class,
    ldr,
    ldr_with_imm,
    ldr_with_imm_stack,
    ldr_with_inc_writeback,
    ldr_with_postinc,
    ldrb_with_imm,
    ldrh_with_imm,
    ldrh_with_postinc,
    ldrb_with_postinc,
    vldr_with_imm,
    vldr_with_postinc,
    Ldrd,
    ldm_interval,
    ldm_interval_inc_writeback,
    vldm_interval_inc_writeback,
    str_with_imm,
    str_with_imm_stack,
    str_with_postinc,
    str_no_off,
    strh_with_imm,
    strh_with_postinc,
    stm_interval_inc_writeback,
    movw_imm,
    movt_imm,
    adds,
    add,
    add_short,
    add_imm,
    add_imm_short,
    mul,
    mul_short,
    smull,
    smlal,
    mla,
    mls,
    smulwb,
    smulwt,
    smultb,
    smultt,
    smulbb,
    smlabt,
    smlabb,
    smlatt,
    smlatb,
    smlad,
    smladx,
    smuad,
    smuadx,
    smmulr,
    neg_short,
    log_and,
    log_or,
    eor,
    eor_short,
    eors,
    eors_short,
    bic,
    bics,
    cmp,
    cmp_imm,
    bne,
    vmov_gpr,
    vmov_gpr2,
    vmov_gpr2_dual,
    pkhbt,
    pkhtb,
    pkhbt_shifted,
    ubfx_imm,
    uadd16,
    usub16,
    sadd16,
    ssub16,
    sub,
    subs_imm,
    subs_imm_short,
    sub_imm_short,
    ror,
    ror_short,
    rors_short,
    lsl,
    asr,
    asrs,
    add_shifted,
    sub_shifted,
    log_and_shifted,
    log_or_shifted,
    eor_shifted,
    bic_shifted,
    Armv7mShiftedArithmetic,
    Armv7mShiftedLogical,
    Armv7mBasicArithmetic,
    Armv7mLogical,
)

from sympy import simplify

issue_rate = 2
llvm_mca_target = "cortex-m7"


class ExecutionUnit(Enum):
    """Enumeration of execution units in Cortex-M7 model"""

    STORE = 0
    ALU0 = 1
    ALU1 = 2
    MAC = 5
    FPU = 6
    LOAD0 = 7
    LOAD1 = 8
    SIMD = 9

    def __repr__(self):
        return self.name

    def ALU():
        return [ExecutionUnit.ALU0, ExecutionUnit.ALU1]

    def SHIFT():
        return [ExecutionUnit.SHIFT0, ExecutionUnit.SHIFT1]

    def LOAD():
        return [ExecutionUnit.LOAD0, ExecutionUnit.LOAD1]


# Opaque function called by SLOTHY to add further microarchitecture-
# specific constraints which are not encapsulated by the general framework.
def add_further_constraints(slothy):
    if slothy.config.constraints.functional_only:
        return
    add_st_hazard(slothy)

    add_dsp_slot_constraint(slothy)
    add_mac_slot_constraint(slothy)


def add_dsp_slot_constraint(slothy):
    slothy.restrict_slots_for_instructions_by_class(
        [pkhbt, pkhtb, pkhbt_shifted, ubfx_imm, uadd16, usub16, sadd16, ssub16], [0]
    )


def add_mac_slot_constraint(slothy):
    slothy.restrict_slots_for_instructions_by_class(
        [
            mul,
            mul_short,
            smull,
            smlal,
            mla,
            mls,
            smulwb,
            smulwt,
            smultb,
            smultt,
            smulbb,
            smlabt,
            smlabb,
            smlatt,
            smlatb,
            smlad,
            smladx,
            smuad,
            smuadx,
            smmulr,
        ],
        [1],
    )


def add_st_hazard(slothy):
    def is_st_ld_pair(inst_a, inst_b):
        return (
            isinstance(inst_a.inst, ldr_with_imm)
            or isinstance(inst_a.inst, ldr_with_imm_stack)
        ) and (
            isinstance(inst_b.inst, str_with_imm)
            or isinstance(inst_b.inst, str_with_imm_stack)
        )

    def evaluate_immediate(string_expr):
        if string_expr is None:
            return 0
        string_expr = str(string_expr)
        return int(simplify(string_expr))

    for t_load, t_store in slothy.get_inst_pairs(cond=is_st_ld_pair):
        if t_load.is_locked and t_store.is_locked:
            continue

        ldr_imm = evaluate_immediate(t_load.inst.immediate)
        str_imm = evaluate_immediate(t_store.inst.immediate)

        if abs(ldr_imm - str_imm) >= 8:
            continue

        ldr_before_str = slothy._NewBoolVar("")
        ldr_after_str = slothy._NewBoolVar("")
        slothy._AddExactlyOne([ldr_before_str, ldr_after_str])
        slothy._Add(t_load.program_start_var < t_store.program_start_var).OnlyEnforceIf(
            ldr_before_str
        )
        slothy._Add(
            t_load.program_start_var >= t_store.program_start_var + 8
        ).OnlyEnforceIf(ldr_after_str)


# Opaque function called by SLOTHY to add further microarchitecture-
# specific objectives.
def has_min_max_objective(config):
    _ = config
    return False


def get_min_max_objective(slothy):
    _ = slothy
    return


execution_units = {
    (
        ldr,
        ldr_with_imm,
        ldr_with_imm_stack,
        ldr_with_inc_writeback,
        ldr_with_postinc,
        ldrb_with_imm,
        ldrh_with_imm,
        ldrh_with_postinc,
        ldrb_with_postinc,
        vldr_with_imm,
        vldr_with_postinc,  # TODO: also FPU?
    ): ExecutionUnit.LOAD(),
    (Ldrd, ldm_interval, ldm_interval_inc_writeback, vldm_interval_inc_writeback): [
        ExecutionUnit.LOAD()
    ],
    (
        str_with_imm,
        str_with_imm_stack,
        str_with_postinc,
        str_no_off,
        strh_with_imm,
        strh_with_postinc,
        stm_interval_inc_writeback,
    ): [[ExecutionUnit.STORE, ExecutionUnit.MAC]],
    (
        movw_imm,
        movt_imm,
        adds,
        add,
        add_short,
        add_imm,
        add_imm_short,
        sub,
        subs_imm,
        subs_imm_short,
        sub_imm_short,
        neg_short,
        log_and,
        log_or,
        eor,
        eor_short,
        eors,
        eors_short,
        bic,
        bics,
        cmp,
        cmp_imm,
        bne,
    ): ExecutionUnit.ALU(),
    (ror, ror_short, rors_short, lsl, asr, asrs): [
        [ExecutionUnit.ALU0],
        [ExecutionUnit.ALU1],
    ],
    (
        mul,
        mul_short,
        smull,
        smlal,
        mla,
        mls,
        smulwb,
        smulwt,
        smultb,
        smultt,
        smulbb,
        smlabt,
        smlabb,
        smlatt,
        smlatb,
        smlad,
        smladx,
        smuad,
        smuadx,
        smmulr,
    ): [ExecutionUnit.MAC],
    (vmov_gpr, vmov_gpr2, vmov_gpr2_dual): [ExecutionUnit.FPU],
    (uadd16, sadd16, usub16, ssub16): list(
        map(list, product(ExecutionUnit.ALU(), [ExecutionUnit.SIMD]))
    ),
    (pkhbt, pkhtb, pkhbt_shifted, ubfx_imm): [[ExecutionUnit.ALU0, ExecutionUnit.SIMD]],
    (Armv7mShiftedArithmetic): [[ExecutionUnit.ALU0]],
    (Armv7mShiftedLogical): [[ExecutionUnit.ALU0]],
}
inverse_throughput = {
    (
        ldr,
        ldr_with_imm,
        ldr_with_imm_stack,
        ldr_with_inc_writeback,
        ldr_with_postinc,
        Ldrd,
        ldrb_with_imm,
        ldrh_with_imm,
        ldrh_with_postinc,
        ldrb_with_postinc,
        vldr_with_imm,
        vldr_with_postinc,  # TODO: double-check
        # actually not, just placeholder
        ldm_interval,
        ldm_interval_inc_writeback,
        vldm_interval_inc_writeback,
        movw_imm,
        movt_imm,
        adds,
        add,
        add_short,
        add_imm,
        add_imm_short,
        add_shifted,
        sub_shifted,
        sub_imm_short,
        subs_imm,
        subs_imm_short,
        uadd16,
        sadd16,
        usub16,
        ssub16,
        mul,
        mul_short,
        smull,
        smlal,
        mla,
        mls,
        smulwb,
        smulwt,
        smultb,
        smultt,
        smulbb,
        smlabt,
        smlabb,
        smlatt,
        smlatb,
        smlad,
        smladx,
        smuad,
        smuadx,
        smmulr,
        neg_short,
        log_and,
        log_and_shifted,
        log_or,
        log_or_shifted,
        eor,
        eor_short,
        eors,
        eors_short,
        eor_shifted,
        bic,
        bics,
        bic_shifted,
        ror,
        ror_short,
        rors_short,
        lsl,
        asr,
        asrs,
        cmp,
        cmp_imm,
        vmov_gpr,
        vmov_gpr2,
        vmov_gpr2_dual,  # verify for dual
        pkhbt,
        pkhtb,
        pkhbt_shifted,
        ubfx_imm,
        str_with_imm,
        str_with_imm_stack,
        str_with_postinc,
        str_no_off,
        strh_with_imm,
        strh_with_postinc,
        bne,
    ): 1,
    (stm_interval_inc_writeback, vmov_gpr2_dual): 2,  # actually not, just placeholder
}

default_latencies = {
    (
        movw_imm,
        movt_imm,
        adds,
        add,
        add_short,
        add_imm,
        add_imm_short,
        add_shifted,
        sub_shifted,
        sub_imm_short,
        subs_imm,
        subs_imm_short,
        uadd16,
        sadd16,
        usub16,
        ssub16,
        neg_short,
        log_and,
        log_and_shifted,
        log_or,
        log_or_shifted,
        eor,
        eor_short,
        eors,
        eors_short,
        bic,
        bics,
        bic_shifted,
        ror,
        ror_short,
        rors_short,
        lsl,
        asr,
        asrs,
        cmp,
        cmp_imm,
        pkhbt,
        pkhtb,
        pkhbt_shifted,
        ubfx_imm,
        vldr_with_imm,
        vldr_with_postinc,  # according to Jan
        # actually not, just placeholder
        ldm_interval,
        ldm_interval_inc_writeback,
        vldm_interval_inc_writeback,
        str_with_imm,
        str_with_imm_stack,
        str_with_postinc,
        str_no_off,
        strh_with_imm,
        strh_with_postinc,
        bne,
    ): 1,
    (
        mul,
        mul_short,
        smull,
        smlal,
        mla,
        mls,
        smulwb,
        smulwt,
        smultb,
        smultt,
        smulbb,
        smlabt,
        smlabb,
        smlatt,
        smlatb,
        smlad,
        smladx,
        smuad,
        smuadx,
        smmulr,
        # TODO: Verify load latency
        stm_interval_inc_writeback,  # actually not, just placeholder
        ldr,
        ldr_with_imm,
        ldr_with_imm_stack,
        ldr_with_inc_writeback,
        ldr_with_postinc,
        ldrb_with_imm,
        ldrh_with_imm,
        ldrh_with_postinc,
        ldrb_with_postinc,
        ldrb_with_postinc,
        eor_shifted,
    ): 2,
    (Ldrd): 3,
    (vmov_gpr2, vmov_gpr2_dual): 3,
    (vmov_gpr): 1,
}


def get_latency(src, out_idx, dst):
    _ = out_idx  # out_idx unused

    instclass_src = find_class(src)
    instclass_dst = find_class(dst)

    latency = lookup_multidict(default_latencies, src, instclass_src)

    # Forwarding path to MAC instructions
    if instclass_dst in [mla, mls, smlabb, smlabt, smlatt, smlatb] and dst.args_in[
        2
    ] in (src.args_out + src.args_in_out):
        latency = latency - 1

    if instclass_dst in [smlal]:
        if len(src.args_out) > 1:
            if (
                src.args_out[0] == dst.args_in_out[0]
                or src.args_out[0] == dst.args_in_out[1]
            ):
                latency = latency - 1
        elif len(src.args_in_out) > 1:
            if (
                src.args_in_out[0] == dst.args_in_out[0]
                or src.args_in_out[0] == dst.args_in_out[1]
            ):
                latency = latency - 1

    # Multiply accumulate chain latency is 1
    if (
        instclass_src in [smlal]
        and instclass_dst in [smlal]
        and src.args_in_out[0] == dst.args_in_out[0]
        and src.args_in_out[1] == dst.args_in_out[1]
    ):
        return 1

    # Load latency is 1 cycle if the destination is an arithmetic/logical instruction
    if (
        instclass_src in [ldr_with_imm, ldr_with_imm_stack, ldr_with_inc_writeback]
        and sum(
            [
                issubclass(instclass_dst, pc)
                for pc in [Armv7mBasicArithmetic, Armv7mLogical]
            ]
        )
        and src.args_out[0] in dst.args_in
    ):
        latency = latency - 1

    # Shifted operand needs to be available one cycle early
    # TODO: verify how this applies to ubfx with imm
    if (
        sum(
            [
                issubclass(instclass_dst, pc)
                for pc in [
                    Armv7mShiftedLogical,
                    Armv7mShiftedArithmetic,
                    pkhbt,
                    pkhtb,
                    pkhbt_shifted,
                ]
            ]
        )
        and dst.args_in[1] in src.args_out
        or sum([issubclass(instclass_dst, pc) for pc in [ubfx_imm]])
        and dst.args_in[0] in src.args_out
    ):
        return latency + 1

    # Load and store multiples take a long time to complete
    if instclass_src in [
        ldm_interval,
        ldm_interval_inc_writeback,
        stm_interval_inc_writeback,
        vldm_interval_inc_writeback,
    ]:
        latency = src.num_out

    # Flag setting -> branch has at least 3 latency
    if (
        instclass_src in [subs_imm, subs_imm_short, cmp, cmp_imm]
        and instclass_dst == bne
    ):
        latency = 2

    # Can always store result in the same cycle
    # TODO: double-check this
    if dst.is_store():
        return 0

    return latency


def get_units(src):
    instclass_src = find_class(src)
    units = lookup_multidict(execution_units, src, instclass_src)

    def evaluate_immediate(string_expr):
        if string_expr is None:
            return 0
        string_expr = str(string_expr)
        return int(simplify(string_expr))

    # The Cortex-M7 has two memory banks
    # If two loads use the same memory bank, they cannot dual issue
    # There are no constraints which load can go to which issue slot
    # Approximiation: Only look at immediates, i.e., assume all pointers are
    # aligned to 8 bytes
    if src.is_ldr():
        imm = evaluate_immediate(src.immediate)

        if (imm % 8) // 4 == 0:
            return [ExecutionUnit.LOAD0]
        else:
            return [ExecutionUnit.LOAD1]

    if isinstance(units, list):
        return units
    return [units]


def get_inverse_throughput(src):
    instclass_src = find_class(src)
    itp = lookup_multidict(inverse_throughput, src, instclass_src)
    if find_class(src) in [
        ldm_interval,
        ldm_interval_inc_writeback,
        stm_interval_inc_writeback,
        vldm_interval_inc_writeback,
    ]:
        itp = src.num_out

    return itp



#
# Copyright (c) 2022 Arm Limited
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Hanno Becker <hannobecker@posteo.de>
#

#
# Playground for local microarchitectural explorations
#
# The baseline is just a copy of the Cortex-M55 model without ST-LD hazard
#

from enum import Enum
from slothy.helper import lookup_multidict
from slothy.targets.arm_v81m.arch_v81m import (
    Instruction,
    nop,
    ldr,
    ldr_with_writeback,
    ldr_with_post,
    mov_imm,
    mvn_imm,
    mov,
    add,
    sub,
    pkhbt,
    add_imm,
    sub_imm,
    vshrnb,
    vshrnt,
    vrshr,
    vrshl,
    vshr,
    vshl,
    vshl_T3,
    vshlc,
    vshllb,
    vshllt,
    vsli,
    vmovlb,
    vmovlt,
    vrev16,
    vrev32,
    vrev64,
    vdup,
    vmov_imm,
    vmov_double_v2r,
    vadd_sv,
    vadd_vv,
    vsub,
    vsub_T2,
    vhadd,
    vhsub,
    vhcadd,
    vand,
    vbic,
    vbic_nodt,
    vorr,
    veor,
    veor_nodt,
    vmulh,
    vmul_T1,
    vmul_T2,
    vmullb,
    vmullt,
    vqrdmulh_T1,
    vqrdmulh_T2,
    vqdmlah,
    vqrdmlah,
    vqdmladhx,
    vqdmlsdh,
    vqdmulh_vv,
    vqdmulh_sv,
    vmla,
    vfma,
    vmulf_T1,
    vmulf_T2,
    ldrd,
    ldrd_no_imm,
    ldrd_with_writeback,
    ldrd_with_post,
    strd,
    strd_with_writeback,
    strd_with_post,
    restored,
    restore,
    saved,
    save,
    qsave,
    qrestore,
    vldrb,
    vldrb_no_imm,
    vldrb_with_writeback,
    vldrb_with_post,
    vldrh,
    vldrh_no_imm,
    vldrh_with_writeback,
    vldrh_with_post,
    vldrw,
    vldrw_no_imm,
    vldrw_with_writeback,
    vldrw_with_post,
    vldrw_gather,
    vldrw_gather_uxtw,
    vldrb_gather,
    vldrb_gather_uxtw,
    vldrh_gather,
    vldrh_gather_uxtw,
    vld20,
    vld21,
    vld20_with_writeback,
    vld21_with_writeback,
    vld40,
    vld41,
    vld42,
    vld43,
    vld40_with_writeback,
    vld41_with_writeback,
    vld42_with_writeback,
    vld43_with_writeback,
    vstrw,
    vstrw_no_imm,
    vstrw_with_writeback,
    vstrw_with_post,
    vstrw_scatter,
    vstrw_scatter_uxtw,
    vst20,
    vst21,
    vst20_with_writeback,
    vst21_with_writeback,
    vst40,
    vst41,
    vst42,
    vst43,
    vst40_with_writeback,
    vst41_with_writeback,
    vst42_with_writeback,
    vst43_with_writeback,
    vcmul,
    vcmla,
    vcadd,
    vaddf,
    vsubf,
    vsubf_T2,
    vcaddf,
)

issue_rate = 1


class ExecutionUnit(Enum):
    SCALAR = (0,)
    LOAD = (1,)
    STORE = (1,)
    VEC_INT = (2,)
    VEC_MUL = (3,)
    VEC_FPU = (3,)
    STACK = (1,)

    def __repr__(self):
        return self.name


# Opaque function called by SLOTHY to add further microarchitecture-
# specific constraints which are not encapsulated by the general framework.
def add_further_constraints(slothy):
    # See ST-LD hazard in M55 model for an example
    return


def has_min_max_objective(slothy):
    # See minimization of ST-LD hazards in M55 model for an example
    return False


def get_min_max_objective(slothy):
    return


execution_units = {
    nop: ExecutionUnit.SCALAR,
    ldr: ExecutionUnit.LOAD,
    ldr_with_writeback: ExecutionUnit.LOAD,
    ldr_with_post: ExecutionUnit.LOAD,
    mov_imm: ExecutionUnit.SCALAR,
    mvn_imm: ExecutionUnit.SCALAR,
    mov: ExecutionUnit.SCALAR,
    add: ExecutionUnit.SCALAR,
    sub: ExecutionUnit.SCALAR,
    pkhbt: ExecutionUnit.SCALAR,
    add_imm: ExecutionUnit.SCALAR,
    sub_imm: ExecutionUnit.SCALAR,
    vshrnb: ExecutionUnit.VEC_INT,
    vshrnt: ExecutionUnit.VEC_INT,
    vrshr: ExecutionUnit.VEC_INT,
    vrshl: ExecutionUnit.VEC_INT,
    vshr: ExecutionUnit.VEC_INT,
    vshl: ExecutionUnit.VEC_INT,
    vshl_T3: ExecutionUnit.VEC_INT,
    vshlc: ExecutionUnit.VEC_INT,
    vshllb: ExecutionUnit.VEC_INT,
    vshllt: ExecutionUnit.VEC_INT,
    vsli: ExecutionUnit.VEC_INT,
    vmovlb: ExecutionUnit.VEC_INT,
    vmovlt: ExecutionUnit.VEC_INT,
    vrev16: ExecutionUnit.VEC_INT,
    vrev32: ExecutionUnit.VEC_INT,
    vrev64: ExecutionUnit.VEC_INT,
    vdup: ExecutionUnit.VEC_INT,
    vmov_imm: [ExecutionUnit.VEC_INT, ExecutionUnit.VEC_MUL],
    vmov_double_v2r: [ExecutionUnit.VEC_INT, ExecutionUnit.VEC_MUL],
    vadd_sv: ExecutionUnit.VEC_INT,
    vadd_vv: ExecutionUnit.VEC_INT,
    vsub: ExecutionUnit.VEC_INT,
    vsub_T2: ExecutionUnit.VEC_INT,
    vhadd: ExecutionUnit.VEC_INT,
    vhsub: ExecutionUnit.VEC_INT,
    vhcadd: ExecutionUnit.VEC_INT,
    vand: ExecutionUnit.VEC_INT,
    vbic: ExecutionUnit.VEC_INT,
    vbic_nodt: ExecutionUnit.VEC_INT,
    vorr: ExecutionUnit.VEC_INT,
    veor: ExecutionUnit.VEC_INT,
    veor_nodt: ExecutionUnit.VEC_INT,
    vmulh: ExecutionUnit.VEC_MUL,
    vmul_T1: ExecutionUnit.VEC_MUL,
    vmul_T2: ExecutionUnit.VEC_MUL,
    vmullb: ExecutionUnit.VEC_MUL,
    vmullt: ExecutionUnit.VEC_MUL,
    vqrdmulh_T1: ExecutionUnit.VEC_MUL,
    vqrdmulh_T2: ExecutionUnit.VEC_MUL,
    vqdmlah: ExecutionUnit.VEC_MUL,
    vqrdmlah: ExecutionUnit.VEC_MUL,
    vqdmladhx: ExecutionUnit.VEC_MUL,
    vqdmlsdh: ExecutionUnit.VEC_MUL,
    vqdmulh_vv: ExecutionUnit.VEC_MUL,
    vqdmulh_sv: ExecutionUnit.VEC_MUL,
    vmla: ExecutionUnit.VEC_MUL,
    vfma: ExecutionUnit.VEC_FPU,
    vmulf_T1: ExecutionUnit.VEC_FPU,
    vmulf_T2: ExecutionUnit.VEC_FPU,
    ldrd: ExecutionUnit.LOAD,
    ldrd_no_imm: ExecutionUnit.LOAD,
    ldrd_with_writeback: ExecutionUnit.LOAD,
    ldrd_with_post: ExecutionUnit.LOAD,
    strd: ExecutionUnit.STORE,
    strd_with_writeback: ExecutionUnit.STORE,
    strd_with_post: ExecutionUnit.STORE,
    restored: ExecutionUnit.STACK,
    restore: ExecutionUnit.STACK,
    saved: ExecutionUnit.STACK,
    save: ExecutionUnit.STACK,
    qsave: ExecutionUnit.STACK,
    qrestore: ExecutionUnit.STACK,
    vldrb: ExecutionUnit.LOAD,
    vldrb_no_imm: ExecutionUnit.LOAD,
    vldrb_with_writeback: ExecutionUnit.LOAD,
    vldrb_with_post: ExecutionUnit.LOAD,
    vldrh: ExecutionUnit.LOAD,
    vldrh_no_imm: ExecutionUnit.LOAD,
    vldrh_with_writeback: ExecutionUnit.LOAD,
    vldrh_with_post: ExecutionUnit.LOAD,
    vldrw: ExecutionUnit.LOAD,
    vldrw_no_imm: ExecutionUnit.LOAD,
    vldrw_with_writeback: ExecutionUnit.LOAD,
    vldrw_with_post: ExecutionUnit.LOAD,
    vldrw_gather: ExecutionUnit.LOAD,
    vldrw_gather_uxtw: ExecutionUnit.LOAD,
    vldrb_gather: ExecutionUnit.LOAD,
    vldrb_gather_uxtw: ExecutionUnit.LOAD,
    vldrh_gather: ExecutionUnit.LOAD,
    vldrh_gather_uxtw: ExecutionUnit.LOAD,
    vld20: ExecutionUnit.LOAD,
    vld21: ExecutionUnit.LOAD,
    vld20_with_writeback: ExecutionUnit.LOAD,
    vld21_with_writeback: ExecutionUnit.LOAD,
    vld40: ExecutionUnit.LOAD,
    vld41: ExecutionUnit.LOAD,
    vld42: ExecutionUnit.LOAD,
    vld43: ExecutionUnit.LOAD,
    vld40_with_writeback: ExecutionUnit.LOAD,
    vld41_with_writeback: ExecutionUnit.LOAD,
    vld42_with_writeback: ExecutionUnit.LOAD,
    vld43_with_writeback: ExecutionUnit.LOAD,
    vstrw: ExecutionUnit.STORE,
    vstrw_no_imm: ExecutionUnit.STORE,
    vstrw_with_writeback: ExecutionUnit.STORE,
    vstrw_with_post: ExecutionUnit.STORE,
    vstrw_scatter: ExecutionUnit.STORE,
    vstrw_scatter_uxtw: ExecutionUnit.STORE,
    vst20: ExecutionUnit.STORE,
    vst21: ExecutionUnit.STORE,
    vst20_with_writeback: ExecutionUnit.STORE,
    vst21_with_writeback: ExecutionUnit.STORE,
    vst40: ExecutionUnit.STORE,
    vst41: ExecutionUnit.STORE,
    vst42: ExecutionUnit.STORE,
    vst43: ExecutionUnit.STORE,
    vst40_with_writeback: ExecutionUnit.STORE,
    vst41_with_writeback: ExecutionUnit.STORE,
    vst42_with_writeback: ExecutionUnit.STORE,
    vst43_with_writeback: ExecutionUnit.STORE,
    vcmul: ExecutionUnit.VEC_FPU,
    vcmla: ExecutionUnit.VEC_FPU,
    vcadd: ExecutionUnit.VEC_FPU,
    vaddf: ExecutionUnit.VEC_FPU,
    vsubf: ExecutionUnit.VEC_FPU,
    vsubf_T2: ExecutionUnit.VEC_FPU,
    vcaddf: ExecutionUnit.VEC_FPU,
}

inverse_throughput = {
    (
        nop,
        mov_imm,
        mvn_imm,
        mov,
        add,
        sub,
        pkhbt,
        add_imm,
        sub_imm,
        vmov_imm,
        vmov_double_v2r,
        ldr,
        ldr_with_writeback,
        ldr_with_post,
        ldrd,
        ldrd_no_imm,
        ldrd_with_writeback,
        ldrd_with_post,
        strd,
        strd_with_writeback,
        strd_with_post,
        restored,
        restore,
        saved,
        save,
    ): 1,
    (
        vrshr,
        vrshl,
        vshrnb,
        vshrnt,
        vdup,
        vshr,
        vshl,
        vshl_T3,
        vshlc,
        vshllb,
        vshllt,
        vsli,
        vmovlb,
        vmovlt,
        vrev16,
        vrev32,
        vrev64,
        vadd_sv,
        vadd_vv,
        vsub,
        vsub_T2,
        vhadd,
        vhsub,
        vhcadd,
        vand,
        vbic,
        vbic_nodt,
        vorr,
        veor,
        veor_nodt,
        vmulh,
        vmul_T1,
        vmul_T2,
        vmullb,
        vmullt,
        vqrdmulh_T1,
        vqrdmulh_T2,
        vqdmlah,
        vqrdmlah,
        vqdmulh_sv,
        vqdmulh_vv,
        vqdmladhx,
        vqdmlsdh,
        vmla,
        vstrw,
        vstrw_no_imm,
        vstrw_with_writeback,
        vstrw_with_post,
        vstrw_scatter,
        vstrw_scatter_uxtw,
        qsave,
        qrestore,
        vldrb,
        vldrb_no_imm,
        vldrb_with_writeback,
        vldrb_with_post,
        vldrh,
        vldrh_no_imm,
        vldrh_with_writeback,
        vldrh_with_post,
        vldrw,
        vldrw_no_imm,
        vldrw_with_writeback,
        vldrw_with_post,
        vldrw_gather,
        vldrw_gather_uxtw,
        vldrb_gather,
        vldrb_gather_uxtw,
        vldrh_gather,
        vldrh_gather_uxtw,
        vld20,
        vld21,
        vld20_with_writeback,
        vld21_with_writeback,
        vld40,
        vld41,
        vld42,
        vld43,
        vld40_with_writeback,
        vld41_with_writeback,
        vld42_with_writeback,
        vld43_with_writeback,
        vst20,
        vst21,
        vst20_with_writeback,
        vst21_with_writeback,
        vst40,
        vst41,
        vst42,
        vst43,
        vst40_with_writeback,
        vst41_with_writeback,
        vst42_with_writeback,
        vst43_with_writeback,
        vcmul,
        vcmla,
        vcadd,
        vaddf,
        vsubf,
        vsubf_T2,
        vhcadd,
        vmulf_T1,
        vmulf_T2,
        vfma,
    ): 2,
}

default_latencies = {
    (
        ldrd,
        ldrd_no_imm,
        ldrd_with_writeback,
        ldrd_with_post,
    ): 2,
    restored: 2,
    (
        ldr,
        ldr_with_writeback,
        ldr_with_post,
        mov_imm,
        mvn_imm,
        mov,
        add,
        sub,
        pkhbt,
        add_imm,
        sub_imm,
        vshr,
        vshl,
        vshl_T3,
        vshlc,
        vrev16,
        vrev32,
        vrev64,
        vdup,
        vmov_imm,
        vmov_double_v2r,
        vadd_vv,
        vadd_sv,
        vsub,
        vsub_T2,
        vhadd,
        vhsub,
        vhcadd,
        vand,
        vbic,
        vbic_nodt,
        vorr,
        veor,
        veor_nodt,
        qsave,
        save,
        qrestore,
        restore,
        vldrb,
        vldrb_no_imm,
        vldrb_with_writeback,
        vldrb_with_post,
        vldrh,
        vldrh_no_imm,
        vldrh_with_writeback,
        vldrh_with_post,
        vldrw,
        vldrw_no_imm,
        vldrw_with_writeback,
        vldrw_with_post,
        vldrw_gather,
        vldrw_gather_uxtw,
        vldrb_gather,
        vldrb_gather_uxtw,
        vldrh_gather,
        vldrh_gather_uxtw,
        vld20,
        vld21,
        vld20_with_writeback,
        vld21_with_writeback,
        vld40,
        vld41,
        vld42,
        vld43,
        vld40_with_writeback,
        vld41_with_writeback,
        vld42_with_writeback,
        vld43_with_writeback,
        vst20,
        vst21,
        vst20_with_writeback,
        vst21_with_writeback,
        vst40,
        vst41,
        vst42,
        vst43,
        vst40_with_writeback,
        vst41_with_writeback,
        vst42_with_writeback,
        vst43_with_writeback,
        vstrw,
        vstrw_no_imm,
        vstrw_with_writeback,
        vstrw_with_post,
        vstrw_scatter,
        vstrw_scatter_uxtw,
    ): 1,
    (
        vrshr,
        vrshl,
        vshrnb,
        vshrnt,
        vshllb,
        vshllt,
        vsli,
        vmovlb,
        vmovlt,
        vmulh,
        vmul_T1,
        vmul_T2,
        vmullb,
        vmullt,
        vqrdmulh_T1,
        vqrdmulh_T2,
        vqdmlah,
        vqrdmlah,
        vqdmulh_sv,
        vqdmulh_vv,
        vqdmladhx,
        vqdmlsdh,
        vmla,
        vcmul,
        vcmla,
        vcadd,
        vaddf,
        vsubf,
        vsubf_T2,
        vhcadd,
        vmulf_T1,
        vmulf_T2,
        vfma,
    ): 2,
}


def find_class(src):
    for inst_class in Instruction.__subclasses__():
        if isinstance(src, inst_class):
            return inst_class
    raise Exception("Couldn't find instruction class")


def get_latency(src, out_idx, dst):
    instclass_src = find_class(src)
    instclass_dst = find_class(dst)

    default_latency = lookup_multidict(default_latencies, src, instclass_src)

    #
    # Check for latency exceptions
    #

    # VMULx -> VSTR has single cycle latency
    if instclass_dst in [
        vstrw,
        vstrw_no_imm,
        vstrw_with_writeback,
        vstrw_with_post,
        vstrw_scatter,
        vstrw_scatter_uxtw,
        qsave,
    ] and instclass_src in [
        vmul_T1,
        vmul_T2,
        vmullb,
        vmullt,
        vqrdmulh_T1,
        vqrdmulh_T2,
        vqdmlah,
        vqrdmlah,
        vqdmulh_vv,
        vqdmulh_sv,
        vmla,
        vqdmladhx,
        vqdmlsdh,
        vcmul,
        vcmla,
        vcadd,
        vaddf,
        vsubf,
        vsubf_T2,
        vfma,
        vmulf_T1,
        vmulf_T2,
        vcaddf,
    ]:
        return 1

    # Inputs to VST4x seem to have higher latency
    # Use 3 cycles as an upper bound here.
    if (
        instclass_dst
        in [
            vst40,
            vst40_with_writeback,
            vst41,
            vst41_with_writeback,
            vst42,
            vst42_with_writeback,
        ]
        or instclass_dst in [vst20, vst20_with_writeback]
    ) and instclass_src in [
        vshr,
        vshl,
        vshl_T3,
        vshlc,
        vrev16,
        vrev32,
        vrev64,
        vdup,
        vmov_imm,
        vadd_vv,
        vadd_sv,
        vsub,
        vsub_T2,
        vhadd,
        vhsub,
        vhcadd,
        vand,
        vbic,
        vbic_nodt,
        vorr,
        veor,
        veor_nodt,
        vrshr,
        vshrnb,
        vshrnt,
        vrshl,
        vshllt,
        vshllt,
        vsli,
        vmovlb,
        vmovlt,
        vmulh,
        vmul_T1,
        vmul_T2,
        vmullb,
        vmullt,
        vqrdmulh_T1,
        vqrdmulh_T2,
        vqdmlah,
        vqrdmlah,
        vqdmulh_sv,
        vqdmulh_vv,
        vqdmladhx,
        vqdmlsdh,
        vmla,
        vfma,
        vmulf_T1,
        vmulf_T2,
        vcmul,
        vcmla,
        vcadd,
        vaddf,
        vsubf,
        vsubf_T2,
        vhcadd,
    ]:
        return 3

    return default_latency


def get_units(src):
    instclass = find_class(src)
    units = execution_units[instclass]
    if isinstance(units, list):
        return units
    else:
        return [units]


def get_inverse_throughput(src):
    instclass = find_class(src)
    return lookup_multidict(inverse_throughput, src, instclass)



#
# Copyright (c) 2022 Arm Limited
# Copyright (c) 2022 Hanno Becker
# Copyright (c) 2023 Amin Abdulrahman, Matthias Kannwischer
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Hanno Becker <hannobecker@posteo.de>
#

import re
import subprocess
import os
import platform
import logging
import inspect
from abc import ABC, abstractmethod
from sympy import simplify
from slothy.targets.common import FatalParsingException, UnknownInstruction

from unicorn import Uc, UcError

# Markers for /* */ comment preservation
_BLOCK_COMMENT_MARKER = "__SLOTHY_BLOCK__"
_NEWLINE_MARKER = "__SLOTHY_NL__"


class SourceLine:
    """Representation of a single line of source code"""

    def _extract_comments_from_text(self):
        if "//" not in self._raw:
            return
        # Don't split block comments (they contain markers and may have // in content)
        if _NEWLINE_MARKER in self._raw:
            s = self._raw.split("//", 1)
        else:
            s = self._raw.split("//")
        self._raw = s[0]
        # Preserve whitespace for block comments (with markers), lstrip others
        self._comments += [c if _NEWLINE_MARKER in c else c.lstrip() for c in s[1:]]
        self._trim_comments()

    def _extract_indentation_from_text(self):
        old = self._raw
        new = old.lstrip()
        self._indentation += len(old) - len(new)
        self._raw = new

    @staticmethod
    def _parse_tags_in_string(s, tags):
        def parse_value(v):
            if v.lower() == "true":
                return True
            if v.lower() == "false":
                return False
            if v.isnumeric():
                return int(v)
            return v

        def tag_value_callback(g):
            tag = g.group("tag")
            value = parse_value(g.group("value"))
            tags[tag] = value
            return ""

        def tag_list_callback(g):
            tag = g.group("tag")
            values = list(map(parse_value, g.group("value").split(",")))
            tags[tag] = values
            return ""

        def tag_callback(g):
            tag = g.group("tag")
            tags[tag] = True
            return ""

        tag_value_regexp_txt = r"@slothy:(?P<tag>(\w|-)+)=(?P<value>[a-zA-Z_0-9\\()]+)"
        tag_list_regexp_txt = r"@slothy:(?P<tag>(\w|-)+)=\[(?P<value>.+)\]"
        tag_regexp_txt = r"@slothy:(?P<tag>(\w|-)+)"
        s = re.sub(tag_value_regexp_txt, tag_value_callback, s)
        s = re.sub(tag_list_regexp_txt, tag_list_callback, s)
        s = re.sub(tag_regexp_txt, tag_callback, s)
        return s

    def _strip_comments(self):
        # Preserve whitespace for block comments (with markers), lstrip others
        self._comments = [
            c if _NEWLINE_MARKER in c else c.lstrip() for c in self._comments
        ]

    def _trim_comments(self):
        self._strip_comments()
        self._comments = list(filter(lambda s: s != "", self._comments))

    def _extract_tags_from_comments(self):
        tags = {}
        self._comments = list(
            map(lambda c: SourceLine._parse_tags_in_string(c, tags), self._comments)
        )
        self._trim_comments()
        self.add_tags(tags)

    def reduce(self):
        """Extract metadata (tags, comments, indentation) from raw text

        The extracted components get retracted from the text."""
        self._extract_indentation_from_text()
        self._extract_comments_from_text()
        self._extract_tags_from_comments()
        return self

    def unify(self):
        """Unify the notation of the input source.
        Replaces tabs by spaces.

        The original text is overwritten."""
        old = self._raw
        new = old.replace("\t", " ")
        self._raw = new

    def add_comment(self, comment):
        """Add a comment to the metadata of a source line"""
        self._comments.append(comment)
        return self

    def add_comments(self, comments):
        """Add one or more comments to the metadata of a source line"""
        for c in comments:
            self.add_comment(c)
        return self

    def set_comments(self, comments):
        """Set comments for source line.

        Overwrites existing comments."""
        self._comments = comments
        return self

    def set_comment(self, comment):
        """Set single comment for source line.

        Overwrites existing comments."""
        self.set_comments([comment])
        return self

    def __init__(self, s, reduce=True):
        assert isinstance(s, str)

        self._raw = s
        self._tags = {}
        self._indentation = 0
        self._fixlength = None
        self._comments = []

        if reduce is True:
            self.reduce()

    def set_tag(self, tag, value=True):
        """Set source line tag"""
        self._tags[tag] = value
        return self

    def set_length(self, length):
        """Set the padded length of the text component of the source line

        When printing the source line with to_string(), the source text will be
        whitespace padded to the specified length before adding comments and tags.
        This allows to print multiple commented source lines with a uniform
        indentation for the comments, improving readability."""
        self._fixlength = length
        return self

    @property
    def tags(self):
        """Return the list of tags for the source line

        Tags are source annotations of the form @slothy:(tag[=value]?).
        """
        return self._tags

    @tags.setter
    def tags(self, v):
        self._tags = v

    @property
    def comments(self):
        """Return the list of comments for the source line"""
        return self._comments

    @comments.setter
    def comments(self, v):
        self._comments = v

    def has_text(self):
        """Indicates if the source line constaints some text"""
        return self._raw.strip() != ""

    @staticmethod
    def instruction_count(lines):
        """Count the number of instructions in a list of source lines"""
        return len([line for line in lines if line.has_text()])

    @property
    def indentation(self):
        """Returns the current level of indentation for the source line"""
        return self._indentation

    @property
    def text(self):
        """Returns the (non-metadata) text in the source line"""
        return self._raw

    def to_string(self, indentation=True, comments=True, tags=True):
        """Convert source line to a string

        This includes formatting the metadata in a way reversing the
        parsing done in the _extract_xxx() routines."""
        if self._fixlength is None:
            core = self._raw
        else:
            core = f"{self._raw:{self._fixlength}s}"

        indentation = " " * self._indentation if indentation is True else ""

        double_comments = filter(lambda t: not t.startswith("/"), self._comments)
        triple_comments = map(
            lambda s: (" " + s[1:].strip()).rstrip(),
            filter(lambda t: t.startswith("/"), self._comments),
        )

        additional = []

        if comments is True:
            # Convert comments with markers back to /* */ format
            def format_comment(s):
                if _NEWLINE_MARKER in s:
                    # Restore as multi-line /* */ comment
                    lines = s.split(_NEWLINE_MARKER)
                    return "/*" + "\n".join(lines) + "*/"
                elif s.startswith(_BLOCK_COMMENT_MARKER):
                    # Restore as single-line /* */ comment
                    return f"/* {s[len(_BLOCK_COMMENT_MARKER):]} */"
                else:
                    # Regular // comment - keep as is
                    return f"// {s}"

            additional += list(map(format_comment, double_comments))
            additional += list(map(lambda s: f"///{s}", triple_comments))

        if tags is True:

            def print_tag_value(tv):
                t, v = tv
                if v is True:
                    return f"// @slothy:{t}"
                #                if isinstance(v, list):
                #                    return f"// @slothy:{t}=[{','.join(v)}]"
                return f"// @slothy:{t}={v}"

            additional += list(map(print_tag_value, self._tags.items()))

        add_str = " ".join(additional)

        return f"{indentation}{core}{add_str}".rstrip()

    def __str__(self):
        raise AsmHelperException(
            "Forbid str(SourceLine) for now -- call SourceLine.to_string() "
            "explicitly and indicate if indentation, comments and tags "
            "should be printed as well."
        )

    @staticmethod
    def reduce_source(src):
        """Extract metadata (e.g. indentation, tags, comments) from source lines"""
        assert SourceLine.is_source(src)
        for line in src:
            line.reduce()
        return [
            line
            for line in src
            if line.has_text()
            and not AsmHelper.is_alignment_directive(line)
            and not AsmHelper.is_allocation_directive(line)
        ]

    @staticmethod
    def unify_source(src):
        """Unify source in source lines."""
        assert SourceLine.is_source(src)
        for line in src:
            line.unify()
        return [line for line in src]

    @staticmethod
    def log(name, s, logger=None, err=False):
        """Send source to logger"""
        assert isinstance(s, list)
        if err:
            fun = logger.error
        else:
            fun = logger.debug
        if len(s) == 0:
            return
        fun(f"Dump: {name}")
        for line in s:
            fun(f"> {line.to_string()}")

    def set_text(self, s):
        """Set the text of the source line

        This only affects the instruction text of the source line, but leaves
        metadata (such as comments, indentation or tags) unmodified."""
        self._raw = s
        return self

    def transform_text(self, f):
        """Apply transformation f to text of source line."""
        self._raw = f(self._raw)

    def add_text(self, s):
        """Add text to a source line

        This only affects the instruction text of the source line, but leaves
        metadata (such as comments, indentation or tags) unmodified."""
        self._raw += " " + s
        return self

    @property
    def is_escaped(self):
        """Indicates if line text ends with a backslash"""
        return self.text.endswith("\\")

    def remove_escaping(self):
        """Remove escape character at end of line, if present"""
        if not self.is_escaped:
            return self
        self._raw = self._raw[:-1]
        return self

    def __len__(self):
        return len(self._raw)

    def copy(self):
        """Create a copy of a source line"""
        return (
            SourceLine(self._raw)
            .add_tags(self._tags.copy())
            .set_indentation(self._indentation)
            .add_comments(self._comments.copy())
            .set_length(self._fixlength)
        )

    @staticmethod
    def read_multiline(s, reduce=True):
        """Parse multi-line string or array of strings into list of SourceLine
        instances"""
        if isinstance(s, str):
            # Retain newline termination
            terminated_by_newline = len(s) > 0 and s[-1] == "\n"

            # Convert /* */ comments to // style with markers for restoration
            def replace_comment(match):
                content = match.group(1)
                if "\n" in content:
                    # Multi-line /* */ comment - mark newlines
                    content = content.replace("\n", _NEWLINE_MARKER)
                    return f"//{content}" if content.strip() else ""
                else:
                    # Single-line /* */ comment - mark with prefix
                    content = content.strip()
                    return f"// {_BLOCK_COMMENT_MARKER}{content}" if content else ""

            s = re.sub(r"/\*(.*?)\*/", replace_comment, s, flags=re.DOTALL)
            s = s.splitlines()
            if terminated_by_newline:
                s.append("")

        return SourceLine.merge_escaped_lines(
            [SourceLine(line, reduce=reduce) for line in s]
        )

    @staticmethod
    def merge_escaped_lines(s):
        """Merge lines ending in a backslash with subsequent line(s)"""
        assert SourceLine.is_source(s)
        res = []
        cur = None
        for line in s:
            if cur is not None:
                cur.add_text(line.text)
                cur.add_tags(line.tags)
                cur.add_comments(line.comments)
            else:
                cur = line.copy()
            if cur.is_escaped:
                cur.remove_escaping()
            else:
                res.append(cur)
                cur = None
        assert cur is None
        return res

    @staticmethod
    def copy_source(s):
        """Create a copy of a list of source lines"""
        assert SourceLine.is_source(s)
        return [line.copy() for line in s]

    @staticmethod
    def write_multiline(s, comments=True, indentation=True, tags=True):
        """Write source as multiline string"""
        return "\n".join(
            map(
                lambda t: t.to_string(
                    comments=comments, tags=tags, indentation=indentation
                ),
                s,
            )
        )

    def set_indentation(self, indentation):
        """Set the indentation (in number of spaces) to be used in to_string()"""
        self._indentation = indentation
        return self

    def add_tags(self, tags):
        """Add one or more tags to the metadata of the source line

        tags must be a tag:value dictionary."""
        self._tags = {**self._tags, **tags}
        return self

    def add_tag(self, tag, value):
        """Add a single tag-value pair to the metadata of the source line

        If a tag is already specified, it is overwritten."""
        return self.add_tags({tag: value})

    def inherit_tags(self, line):
        """Inherits the tags from another source line

        In case of overlapping tags, source line l takes precedence."""
        assert SourceLine.is_source_line(line)
        self.add_tags(line.tags)
        return self

    def inherit_comments(self, line):
        """Inherits the comments from another source line"""
        assert SourceLine.is_source_line(line)
        self.add_comments(line.comments)
        return self

    @staticmethod
    def apply_indentation(source, indentation):
        """Apply consistent indentation to assembly source"""
        assert SourceLine.is_source(source)
        if indentation is None:
            return source
        assert isinstance(indentation, int)
        return [line.copy().set_indentation(indentation) for line in source]

    @staticmethod
    def drop_tags(source):
        """Drop all tags from a source"""
        assert SourceLine.is_source(source)
        for line in source:
            line.tags = {}
        return source

    @staticmethod
    def split_semicolons(s):
        """ "Split the text of a source line at semicolons

        The resulting source lines inherit their metadata from the caller."""
        assert SourceLine.is_source(s)
        res = []
        for line in s:
            for ll in line.text.split(";"):
                t = line.copy()
                t.set_text(ll)
                res.append(t)
        return res

    @staticmethod
    def is_source(s):
        """Check if parameter is a list of SourceLine instances"""
        if isinstance(s, list) is False:
            return False
        for t in s:
            if isinstance(t, SourceLine) is False:
                return False
        return True

    @staticmethod
    def is_source_line(s):
        """Checks if the parameter is a SourceLine instance"""
        return isinstance(s, SourceLine)


class NestedPrint:
    """Helper for recursive printing of structures"""

    def __str__(self):
        top = [self.__class__.__name__ + ":"]
        res = []
        indent = " " * 8
        for name, value in vars(self).items():
            res += f"{name}: {value}".splitlines()
        res = top + [indent + r for r in res]
        return "\n".join(res)

    def log(self, fun):
        """Pass self-description line-by-line to logging function"""
        for line in str(self).splitlines():
            fun(line)


class LockAttributes:
    """Base class adding support for 'locking' the set of attributes, that is,
    preventing the creation of any further attributes. Note that the modification
    of already existing attributes remains possible.

    Our primary use case is for configurations, where this class is used to catch typos
    in the user configuration."""

    def __init__(self):
        self.__dict__["_locked"] = False
        self._locked = False

    def lock(self):
        """Lock set of attributes"""
        self._locked = True

    def __setattr__(self, attr, val):
        if self._locked and attr not in dir(self):
            varlist = [v for v in dir(self) if not v.startswith("_")]
            varlist = "\n".join(map(lambda x: "* " + x, varlist))
            raise TypeError(
                f"Unknown attribute {attr}. \nValid attributes are:\n{varlist}"
            )
        if self._locked and attr == "_locked":
            raise TypeError("Can't unlock an object")
        object.__setattr__(self, attr, val)


class AsmHelperException(Exception):
    """An exception encountered during an assembly helper"""


class AsmHelper:
    """Some helper functions for dealing with assembly"""

    _REGEXP_ALIGN_TXT = r"^\s*\.(?:p2)?align"
    _REGEXP_ALIGN = re.compile(_REGEXP_ALIGN_TXT)

    @staticmethod
    def find_indentation(source):
        """Attempts to find the prevailing indentation in a piece of assembly"""

        def get_indentation(line):
            return len(line) - len(line.lstrip())

        source = map(SourceLine.to_string, source)

        # Remove empty lines
        source = list(filter(lambda t: t.strip() != "", source))
        le = len(source)

        if le == 0:
            return None

        indentations = list(map(get_indentation, source))

        # Some labels may use a different indentation -- here, we just check if
        # there's a dominant indentation
        top_start = (3 * le) // 4
        indentations.sort()
        indentations = indentations[top_start:]

        if indentations[0] == indentations[-1]:
            return indentations[0]

        return None

    @staticmethod
    def rename_function(source, old_funcname, new_funcname):
        """Rename function in assembly snippet"""

        # For now, just replace function names line by line
        def change_funcname(line):
            s = line.text
            s = re.sub(f"{old_funcname}:", f"{new_funcname}:", s)
            s = re.sub(f"\\.global(\\s+){old_funcname}", f".global\\1{new_funcname}", s)
            s = re.sub(f"\\.type(\\s+){old_funcname}", f".type\\1{new_funcname}", s)
            s = re.sub(
                f"\\.size(\\s+){old_funcname},(\\s*)\\.-{old_funcname}",
                f".size {new_funcname}, .-{new_funcname}",
                s,
            )
            return line.copy().set_text(s)

        return [change_funcname(s) for s in source]

    @staticmethod
    def is_alignment_directive(line):
        """Checks is source line is an alignment directive `.[p2]align _`"""
        assert SourceLine.is_source_line(line)
        return AsmHelper._REGEXP_ALIGN.match(line.text) is not None

    @staticmethod
    def is_allocation_directive(line):
        """Checks is source line is an allocation directive."""
        assert SourceLine.is_source_line(line)
        return AsmAllocation.is_allocation(line) or AsmAllocation.is_deallocation(line)

    @staticmethod
    def extract(source, lbl_start=None, lbl_end=None):
        """Extract code between two labels from an assembly source"""
        pre, body, post = AsmHelper._extract_core(source, lbl_start, lbl_end)
        return pre, body, post

    @staticmethod
    def _extract_core(source, lbl_start=None, lbl_end=None):
        pre = []
        body = []
        post = []

        lines = iter(source)
        if lbl_start is None and lbl_end is None:
            body = source
            return pre, body, post

        loop_lbl_regexp_txt = r"^\s*(?P<label>\w+)\s*:(?P<remainder>.*)$"
        loop_lbl_regexp = re.compile(loop_lbl_regexp_txt)
        line = None
        keep = False
        # 0: haven't found initial label yet, 1: between labels, 2: after snd label
        state = 0

        # If no start label is provided, scan from the start to the end label
        if lbl_start is None:
            state = 1

        idx = 0
        while True:
            idx += 1
            if not keep:
                line = next(lines, None)
            if line is None:
                break
            l_str = line.text
            keep = False
            if state == 2:
                post.append(line)
                continue
            expect_label = [lbl_start, lbl_end][state]
            cur_buf = [pre, body][state]
            p = loop_lbl_regexp.match(l_str)
            if p is not None and p.group("label") == expect_label:
                line = line.copy().set_text(p.group("remainder"))
                keep = True
                state += 1
                continue
            cur_buf.append(line)
            continue

        if state < 2:
            if lbl_start is not None and lbl_end is not None:
                raise AsmHelperException(
                    f"Failed to identify region {lbl_start}-{lbl_end}"
                )
            if state == 0:
                if lbl_start is not None:
                    lbl = lbl_start
                else:
                    lbl = lbl_end
                raise AsmHelperException(f"Couldn't find label {lbl}")

        return pre, body, post


class AsmAllocation:
    """Helper for tracking register aliases via .req and .unreq"""

    # TODO: This is conceptionally different and should be
    # handled in its own class.
    _REGEXP_EQU_TXT = (
        r"\s*\.equ\s+(?P<key>[_a-zA-Z][_a-zA-Z0-9]*)\s*,"
        r"\s*(?P<val>[A-Za-z0-9()*/+\-\_\s]+)"
    )

    _REGEXP_REQ_TXT = r"\s*(?P<alias>\w+)\s+\.req\s+(?P<reg>\w+)"
    _REGEXP_UNREQ_TXT = r"\s*\.unreq\s+(?P<alias>\w+)"

    _REGEXP_EQU = re.compile(_REGEXP_EQU_TXT)
    _REGEXP_REQ = re.compile(_REGEXP_REQ_TXT)
    _REGEXP_UNREQ = re.compile(_REGEXP_UNREQ_TXT)

    def __init__(self):
        self.allocations = {}

    def _add_allocation(self, alias, reg):
        if alias in self.allocations:
            raise AsmHelperException(f"Double definition of alias {alias}")
        if reg in self.allocations:
            reg_name = self.allocations[reg]
        else:
            reg_name = reg
        self.allocations[alias] = reg_name

    def _remove_allocation(self, alias):
        if alias not in self.allocations:
            raise AsmHelperException(
                f"Couldn't find alias {alias} --" " .unreq without .req in your source?"
            )
        del self.allocations[alias]

    @staticmethod
    def check_allocation(line):
        """Check if an assembly line is a .req directive. Return the pair
        of alias and register, if so. Otherwise, return None."""
        assert SourceLine.is_source_line(line)

        p = AsmAllocation._REGEXP_REQ.match(line.text)
        if p is not None:
            alias = p.group("alias")
            reg = p.group("reg")
            return alias, reg

        p = AsmAllocation._REGEXP_EQU.match(line.text)
        if p is not None:
            key = p.group("key")
            val = p.group("val")
            return key, val

        return None

    @staticmethod
    def check_deallocation(line):
        """Check if an assembly line is an .unreq directive. Return
        the deallocated alias, if so. Otherwise, return None."""
        assert SourceLine.is_source_line(line)

        p = AsmAllocation._REGEXP_UNREQ.match(line.text)
        if p is not None:
            alias = p.group("alias")
            return alias

        return None

    @staticmethod
    def is_allocation(line):
        return AsmAllocation.check_allocation(line) is not None

    @staticmethod
    def is_deallocation(line):
        return AsmAllocation.check_deallocation(line) is not None

    def parse_line(self, line):
        """Check if an assembly line is a .req .unreq directive, and update the
        alias dictionary accordingly. Otherwise, do nothing."""
        assert SourceLine.is_source_line(line)

        r = AsmAllocation.check_allocation(line)
        if r is not None:
            alias, reg = r
            self._add_allocation(alias, reg)
            return

        r = AsmAllocation.check_deallocation(line)
        if r is not None:
            alias = r
            self._remove_allocation(alias)
            return

        # We ignore everything else

    def parse(self, src):
        """Build register alias dictionary from assembly source"""
        for s in src:
            self.parse_line(s)

    @staticmethod
    def parse_allocs(src):
        """ "Parse register aliases in assembly source into AsmAllocation object."""
        allocs = AsmAllocation()
        allocs.parse(src)
        return allocs.allocations

    @staticmethod
    def unfold_all_aliases(aliases, src):
        """Unfold aliases in assembly source"""

        def _apply_single_alias_to_line(alias_from, alias_to, src):
            res = re.sub(f"(\\W){alias_from}(\\W|\\Z)", f"\\g<1>{alias_to}\\2", src)
            return res

        def _apply_multiple_aliases_to_line(line):
            do_again = True
            while do_again:
                do_again = False
                for alias_from, alias_to in aliases.items():
                    line_new = _apply_single_alias_to_line(alias_from, alias_to, line)
                    if line_new != line:
                        do_again = True
                    line = line_new
            return line

        res = []
        for line in src:
            t = line.copy()
            t.set_text(_apply_multiple_aliases_to_line(line.text))
            res.append(t)
        return res


class BinarySearchLimitException(Exception):
    """Binary search has exceeded its limit without finding a solution"""


def binary_search(
    func, threshold=256, minimum=-1, start=0, precision=1, timeout_below_precision=None
):
    """Conduct a binary search"""
    start = max(start, minimum)
    last_failure = minimum
    val = start
    # Find _some_ version that works
    while True:
        if val > threshold:
            raise BinarySearchLimitException

        def double_val(val):
            if val == 0:
                return 1
            return 2 * val

        success, result = func(val)
        if success:
            last_success = val
            last_success_core = result
            break
        last_failure = val
        val = double_val(val)
    # Find _first_ version that works
    while last_success - last_failure > 1:
        timeout = None
        if last_success - last_failure <= precision:
            if timeout_below_precision is None:
                break
            timeout = timeout_below_precision
        val = last_failure + (last_success - last_failure) // 2
        success, result = func(val, timeout=timeout)
        if success:
            last_success = val
            last_success_core = result
        else:
            last_failure = val
    return last_success, last_success_core


class AsmMacro:
    """Helper class for parsing and applying assembly macros"""

    def __init__(self, name, args, body):
        self.name = name
        self.args = args
        self.body = body

    def __call__(self, args_dict):

        def prepare_value(a):
            a = a.strip()
            a = a.replace("\\", "\\\\")
            if a.startswith("\\") and "\\\\()" not in a:
                a = a + "\\\\()"
            return a

        def apply_arg(ll, arg, val):
            # This function is also called on the values of tags, which may not be
            # strings.
            if isinstance(ll, str) is False:
                return ll
            ll = re.sub(f"\\\\{arg}\\\\\\(\\)", val, ll)
            ll = re.sub(f"\\\\{arg}(\\W|$)", val + "\\1", ll)
            ll = ll.replace("\\()\\()", "\\()")
            return ll

        def apply_args(ll):
            for arg in self.args:
                val = prepare_value(args_dict[arg])
                if not isinstance(ll, list):
                    ll = apply_arg(ll, arg, val)
                else:
                    ll = list(map(lambda x: apply_arg(x, arg, val), ll))
            return ll

        output = []
        for line in self.body:
            t = line.copy()
            t.transform_text(apply_args)
            t.tags = {k: apply_args(v) for (k, v) in t.tags.items()}
            output.append(t)
        return output

    def __repr__(self):
        return self.name

    def unfold_in(self, source, change_callback=None, inherit_comments=False):
        """Unfold all applications of macro in assembly source"""
        assert SourceLine.is_source(source)

        macro_regexp_txt = rf"^\s*{self.name}"
        arg_regexps = []
        if self.args == [""]:
            while True:
                continue

        if len(self.args) > 0:
            macro_regexp_txt = macro_regexp_txt + "\\s+"

        for arg in self.args:
            arg_regexps.append(rf"\s*(?P<{arg}>[^,]+)\s*")

        macro_regexp_txt += "(,|\\s)".join(arg_regexps)
        macro_regexp = re.compile(macro_regexp_txt)

        output = []

        # Go through source line by line and check if there's a macro invocation
        for line in source:
            assert SourceLine.is_source_line(line)

            if line.has_text():
                p = macro_regexp.match(line.text)
            else:
                p = None

            if p is None:
                output.append(line)
                continue
            if change_callback:
                change_callback()
            # Try to keep indentation
            repl = self(p.groupdict())
            for l0 in repl:
                l0.set_indentation(line.indentation)
                l0.inherit_tags(line)
                if inherit_comments is True:
                    l0.inherit_comments(line)
            output += repl

        return output

    @staticmethod
    def unfold_all_macros(macros, source, **kwargs):
        """Unfold list of macros in assembly source"""
        assert isinstance(macros, list)
        assert SourceLine.is_source(source)

        def list_of_instances(line, c):
            return isinstance(line, list) and all(map(lambda m: isinstance(m, c), line))

        def dict_of_instances(line, c):
            return isinstance(line, dict) and list_of_instances(list(line.values()), c)

        if SourceLine.is_source(macros):
            macros = AsmMacro.extract(macros)
        if not dict_of_instances(macros, AsmMacro):
            raise AsmHelperException(f"Invalid argument: {macros}")

        change = True
        while change:
            change = False

            def cb():
                nonlocal change
                change = True

            for m in macros.values():
                source = m.unfold_in(source, change_callback=cb, **kwargs)
        return source

    @staticmethod
    def extract(source):
        """Parse all macro definitions in assembly source file"""

        macros = {}

        state = 0  # 0: Not in a macro 1: In a macro
        current_macro = None
        current_args = None
        current_body = None

        macro_start_regexp_txt = r"^\s*\.macro\s+(?P<name>\w+)(?P<args>.*)$"
        macro_start_regexp = re.compile(macro_start_regexp_txt)

        macro_end_regexp_txt = r"^\s*\.endm\s*$"
        macro_end_regexp = re.compile(macro_end_regexp_txt)

        for cur in source:
            cur_str = cur.text

            if state == 0:

                p = macro_start_regexp.match(cur_str)
                if p is None:
                    continue

                if cur.tags.get("no-unfold", None) is not None:
                    continue

                current_args = [
                    a.strip()
                    for a in re.split(r"\s|\,", p.group("args"))
                    if a.strip() != ""
                ]
                current_macro = p.group("name")
                current_body = []

                if current_args == [""]:
                    current_args = []

                state = 1
                continue

            if state == 1:
                p = macro_end_regexp.match(cur_str)
                if p is None:
                    current_body.append(cur)
                    continue

                macros[current_macro] = AsmMacro(
                    current_macro, current_args, current_body
                )

                current_macro = None
                current_body = None
                current_args = None

                state = 0
                continue

        return macros

    @staticmethod
    def extract_from_file(filename):
        """Parse all macro definitions in assembly file"""
        with open(filename, "r", encoding="utf-8") as f:
            res = AsmMacro.extract(f.read().splitlines())
        return res


class AsmIfElse:
    _REGEXP_IF_TXT = r"\s*\.if\s+(?P<cond>.*)"
    _REGEXP_ELSE_TXT = r"\s*\.else"
    _REGEXP_ENDIF_TXT = r"\s*\.endif"

    _REGEXP_IF = re.compile(_REGEXP_IF_TXT)
    _REGEXP_ELSE = re.compile(_REGEXP_ELSE_TXT)
    _REGEXP_ENDIF = re.compile(_REGEXP_ENDIF_TXT)

    @staticmethod
    def check_if(line):
        """Check if an assembly line is a .req directive. Return the pair
        of alias and register, if so. Otherwise, return None."""
        assert SourceLine.is_source_line(line)

        p = AsmIfElse._REGEXP_IF.match(line.text)
        if p is not None:
            return p.group("cond")
        return None

    @staticmethod
    def is_if(line):
        return AsmIfElse.check_if(line) is not None

    @staticmethod
    def check_else(line):
        """Check if an assembly line is a .req directive. Return the pair
        of alias and register, if so. Otherwise, return None."""
        assert SourceLine.is_source_line(line)

        p = AsmIfElse._REGEXP_ELSE.match(line.text)
        if p is not None:
            return True
        return None

    @staticmethod
    def is_else(line):
        return AsmIfElse.check_else(line) is not None

    @staticmethod
    def check_endif(line):
        """Check if an assembly line is a .req directive. Return the pair
        of alias and register, if so. Otherwise, return None."""
        assert SourceLine.is_source_line(line)

        p = AsmIfElse._REGEXP_ENDIF.match(line.text)
        if p is not None:
            return True
        return None

    @staticmethod
    def is_endif(line):
        return AsmIfElse.check_endif(line) is not None

    @staticmethod
    def evaluate_condition(condition):
        """Evaluates the condition string and returns True or False."""
        try:
            # Evaluate the condition and return the result.
            return simplify(condition)
        except Exception as e:
            print(f"Error evaluating condition '{condition}': {e}")
            return False

    @staticmethod
    def process_instructions(instructions):
        """Processes a list of instructions with conditional statements."""
        output_lines = []
        skip_stack = []

        for instruction in instructions:
            if AsmIfElse.is_if(instruction):
                # Extract condition and evaluate it.
                condition = AsmIfElse.check_if(instruction)
                if AsmIfElse.evaluate_condition(condition):
                    skip_stack.append(False)
                else:
                    skip_stack.append(True)
                continue
            elif AsmIfElse.is_else(instruction):
                if skip_stack:
                    # Invert the top of the stack
                    skip_stack[-1] = not skip_stack[-1]
                continue  # Skip adding the .else line to output
            elif AsmIfElse.is_endif(instruction):
                if skip_stack:
                    skip_stack.pop()  # Exit the current .if block
                continue  # Skip adding the .endif line to output

            # Determine if the current line should be skipped
            if skip_stack and True in skip_stack:
                continue  # Skip lines when inside a false .if block

            # Add the line to output if not skipped
            output_lines.append(instruction)

        return output_lines


class CPreprocessor:
    """Helper class for the application of the C preprocessor"""

    magic_string_start = "SLOTHY_PREPROCESSED_REGION_BEGIN"
    magic_string_end = "SLOTHY_PREPROCESSED_REGION_END"

    @staticmethod
    def unfold(header, body, post, gcc, include=None):
        """Runs the concatenation of header and body through the preprocessor"""

        assert SourceLine.is_source(body)
        assert SourceLine.is_source(header)
        assert SourceLine.is_source(post)

        body_txt = SourceLine.write_multiline(body)
        header_txt = SourceLine.write_multiline(header)
        footer_txt = SourceLine.write_multiline(post)

        code_txt = "\n".join(
            [
                header_txt,
                CPreprocessor.magic_string_start,
                body_txt,
                CPreprocessor.magic_string_end,
                footer_txt,
            ]
        )

        if include is None:
            include = []
            # Ignore #include's
            code_txt = code_txt.replace("#include", "//#include")
        else:
            include = ["-I", include]

        cmd = [gcc] + include + ["-E", "-CC", "-x", "assembler-with-cpp", "-"]

        # Pass -CC to keep comments
        r = subprocess.run(
            cmd, input=code_txt, text=True, capture_output=True, check=True
        )

        unfolded_code = r.stdout.split("\n")
        magic_idx_start = unfolded_code.index(CPreprocessor.magic_string_start)
        magic_idx_end = unfolded_code.index(CPreprocessor.magic_string_end)
        unfolded_code = unfolded_code[magic_idx_start + 1 : magic_idx_end]

        return [SourceLine(r) for r in unfolded_code]


class LLVM_Mc_Error(Exception):
    """Exception thrown if llvm-mc subprocess fails"""


class LLVM_Mc:
    """Helper class for the application of the LLVM MC tool"""

    @staticmethod
    def llvm_mc_output_extract_text_section(objfile):
        """Extracts offset and size of .text section from an objectfile
        emitted by llvm-mc."""

        # We use llvm-readobj to inspect the objectfile, which works
        # for both ELF and MachOS object files. Unfortunately, however,
        # the output formats of both tools are not the same. Moreovoer,
        # the output when selecting JSON as the output format, is not valid JSON.
        # So we're left to hacky string munging.

        # Feed object file through llvm-readobj
        r = subprocess.run(
            ["llvm-readobj", "-S", "-"], input=objfile, capture_output=True, check=True
        )
        objfile_txt = r.stdout.decode().split("\n")

        # We expect something like this here
        # ```
        # File: test.o
        # Format: Mach-O arm
        # Arch: arm
        # AddressSize: 32bit
        # Sections [
        #   Section {
        #     Index: 0
        #     Name: __text (5F 5F 74 65 78 74 00 00 00 00 00 00 00 00 00 00)
        #     Segment: __TEXT (5F 5F 54 45 58 54 00 00 00 00 00 00 00 00 00 00)
        #     Address: 0x0
        #     Size: 0x4
        #     Offset: 176
        #     Alignment: 0
        #     RelocationOffset: 0x0
        #     RelocationCount: 0
        #     Type: Regular (0x0)
        #     Attributes [ (0x800004)
        #       PureInstructions (0x800000)
        #       SomeInstructions (0x4)
        #     ]
        #     Reserved1: 0x0
        #     Reserved2: 0x0
        #   }
        # ]
        # ```
        # So we look for lines "Name: __text" and lines "Offset: ...".
        def parse_as_int(s):
            if s.startswith("0x"):
                return int(s, base=16)
            else:
                return int(s, base=10)

        sections = filter(lambda line: line.strip().startswith("Name: "), objfile_txt)
        sections = list(
            map(
                lambda line: line.strip().removeprefix("Name: ").split(" ")[0].strip(),
                sections,
            )
        )
        offsets = filter(lambda line: line.strip().startswith("Offset: "), objfile_txt)
        offsets = map(
            lambda line: parse_as_int(line.strip().removeprefix("Offset: ")), offsets
        )
        sizes = filter(lambda line: line.strip().startswith("Size: "), objfile_txt)
        sizes = map(
            lambda line: parse_as_int(line.strip().removeprefix("Size: ")), sizes
        )
        sections_with_offsets = {
            s: (o, sz) for (s, o, sz) in zip(sections, offsets, sizes)
        }
        text_section = list(filter(lambda s: "text" in s, sections))
        if len(text_section) != 1:
            raise LLVM_Mc_Error(
                f"Could not find unambiguous text section in object file. Sections: "
                f"{sections}"
            )
        return sections_with_offsets[text_section[0]]

    @staticmethod
    def llvm_mc_output_extract_symbol(objfile, symbol):
        """Extracts symbol from an objectfile emitted by llvm-mc"""

        # Feed object file through llvm-readobj
        r = subprocess.run(
            ["llvm-readobj", "-s", "-"], input=objfile, capture_output=True, check=True
        )
        objfile_txt = r.stdout.decode().split("\n")

        # So we look for lines "Name: ..." and lines "Value: ...".
        def parse_as_int(s):
            if s.startswith("0x"):
                return int(s, base=16)
            else:
                return int(s, base=10)

        symbols = filter(lambda line: line.strip().startswith("Name: "), objfile_txt)
        symbols = list(
            map(
                lambda line: line.strip().removeprefix("Name: ").split(" ")[0].strip(),
                symbols,
            )
        )
        values = filter(lambda line: line.strip().startswith("Value: "), objfile_txt)
        values = map(
            lambda line: parse_as_int(line.strip().removeprefix("Value: ")), values
        )
        symbols_with_values = {s: val for (s, val) in zip(symbols, values)}
        matching_symbols = list(filter(lambda s: s.endswith(symbol), symbols))
        # Sometimes assemble functions are named both `_foo` and `foo`, in which case
        # we'd find multiple matching symbols -- however, they'd have the same value.
        # Hence, only fail if there are multiple matching symbols of _different_ values.
        if len({symbols_with_values[s] for s in matching_symbols}) != 1:
            raise LLVM_Mc_Error(
                f"Could not find unambiguous symbol {symbol} in object file. "
                f"Symbols: {symbols}"
            )
        return symbols_with_values[matching_symbols[0]]

    @staticmethod
    def assemble(
        source, arch, attr, log, symbol=None, preprocessor=None, include_paths=None
    ):
        """Runs LLVM-MC tool to assemble `source`, returning byte code"""

        thumb = "thumb" in arch or (attr is not None and "thumb" in attr)

        # Unfortunately, there is no option to directly extract byte code
        # from LLVM-MC: One either gets a textual description, or an object file.
        # To not introduce another binary dependency, we just extract the byte
        # code directly from the textual output, which for every assembly line
        # has a "encoding: [byte0, byte1, ...]" comment at the end.

        if symbol is None:
            if thumb is True:
                source = [SourceLine(".thumb")] + source
            source = [
                SourceLine(".global harness"),
                SourceLine(".type harness, %function"),
                SourceLine("harness:"),
            ] + source
            symbol = "harness"

        if preprocessor is not None:
            # First, run the C preprocessor on the code
            try:
                source = CPreprocessor.unfold(
                    [], source, [], preprocessor, include=include_paths
                )
            except subprocess.CalledProcessError as exc:
                log.error("CPreprocessor failed on the following input")
                log.error(SourceLine.write_multiline(source))
                raise LLVM_Mc_Error from exc

        if platform.system() == "Darwin":
            source = list(
                filter(lambda s: s.text.strip().startswith(".type") is False, source)
            )

        code = SourceLine.write_multiline(source)

        log.debug("Calling LLVM MC assmelber on the following code")
        log.debug(code)

        args = [f"--arch={arch}", "--assemble", "--filetype=obj"]
        if attr is not None:
            args.append(f"--mattr={attr}")
        try:
            r = subprocess.run(
                ["llvm-mc"] + args, input=code.encode(), capture_output=True, check=True
            )
        except subprocess.CalledProcessError as exc:
            log.error("llvm-mc failed to handle the following code")
            log.error(code)
            log.error("Output from llvm-mc")
            log.error(exc.stderr.decode())
            raise LLVM_Mc_Error from exc

        # TODO: If there are relocations remaining, we should fail at this point

        objfile = r.stdout
        offset, sz = LLVM_Mc.llvm_mc_output_extract_text_section(objfile)
        code = objfile[offset : offset + sz]

        offset = LLVM_Mc.llvm_mc_output_extract_symbol(objfile, symbol)

        if platform.system() == "Darwin" and thumb is True:
            offset += 1

        return code, offset


class LLVM_Mca_Error(Exception):
    """Exception thrown if llvm-mca subprocess fails"""


class LLVM_Mca:
    """Helper class for the application of the LLVM MCA tool"""

    @staticmethod
    def run(header, body, arch, cpu, log, full=False, issue_width=None):
        """Runs LLVM-MCA tool on body and returns result as array of strings"""

        LLVM_MCA_BEGIN = SourceLine("").add_comment("LLVM-MCA-BEGIN")
        LLVM_MCA_END = SourceLine("").add_comment("LLVM-MCA-END")
        mca_binary = "llvm-mca"

        data = SourceLine.write_multiline(
            header + [LLVM_MCA_BEGIN] + body + [LLVM_MCA_END]
        )

        try:
            if full is False:
                args = [
                    "--instruction-info=0",
                    "--dispatch-stats=0",
                    "--timeline=1",
                    "--timeline-max-cycles=0",
                    "--timeline-max-iterations=3",
                ]
            else:
                args = [
                    "--all-stats",
                    "--all-views",
                    "--bottleneck-analysis",
                    "--timeline=1",
                    "--timeline-max-cycles=0",
                    "--timeline-max-iterations=3",
                ]
            if issue_width is not None:
                args += ["--dispatch", str(issue_width)]
            r = subprocess.run(
                [mca_binary, f"--mcpu={cpu}", f"--march={arch}"] + args,
                input=data,
                text=True,
                capture_output=True,
                check=True,
            )
        except subprocess.CalledProcessError as exc:
            raise LLVM_Mca_Error from exc
        res = r.stdout.split("\n")
        return res


class SelfTestException(Exception):
    """Exception thrown upon selftest failures"""


class SelfTest:

    @staticmethod
    def run(
        config,
        log,
        codeA,
        codeB,
        address_registers,
        output_registers,
        iterations,
        fnsym=None,
    ):
        CODE_BASE = 0x010000
        CODE_SZ = 0x010000
        CODE_END = CODE_BASE + CODE_SZ
        RAM_BASE = 0x030000
        RAM_SZ = 0x010000
        STACK_BASE = 0x040000
        STACK_SZ = 0x010000
        STACK_TOP = STACK_BASE + STACK_SZ

        regs = [
            r
            for ty in config.arch.RegisterType
            for r in config.arch.RegisterType.list_registers(ty)
        ]

        def run_code(code, txt=None):
            objcode, offset = LLVM_Mc.assemble(
                code,
                config.arch.llvm_mc_arch,
                config.arch.llvm_mc_attr,
                log,
                symbol=fnsym,
                preprocessor=config.compiler_binary,
                include_paths=config.compiler_include_paths,
            )
            # Setup emulator
            mu = Uc(config.arch.unicorn_arch, config.arch.unicorn_mode)
            # Copy initial register contents into emulator
            for r, v in initial_register_contents.items():
                ur = config.arch.RegisterType.unicorn_reg_by_name(r)
                if ur is None:
                    continue
                mu.reg_write(ur, v)
            if fnsym is not None:
                # If we expect a function return, put a valid address in the LR
                # that serves as the marker to terminate emulation
                mu.reg_write(config.arch.RegisterType.unicorn_link_register(), CODE_END)
            # Setup stack and allocate initial stack memory
            mu.reg_write(
                config.arch.RegisterType.unicorn_stack_pointer(),
                STACK_TOP - config.selftest_default_memory_size,
            )
            # Copy code into emulator
            mu.mem_map(CODE_BASE, CODE_SZ)
            mu.mem_write(CODE_BASE, objcode)

            # Copy initial memory contents into emulator
            mu.mem_map(RAM_BASE, RAM_SZ)
            mu.mem_write(RAM_BASE, initial_memory)
            # Setup stack
            mu.mem_map(STACK_BASE, STACK_SZ)
            mu.mem_write(STACK_BASE, initial_stack)
            # Run emulator
            try:
                # For a function, expect a function return; otherwise, expect
                # to run to the address CODE_END stored in the link register
                if fnsym is None:
                    mu.emu_start(CODE_BASE + offset, CODE_BASE + len(objcode))
                else:
                    mu.emu_start(CODE_BASE + offset, CODE_END)
            except UcError as e:
                log.error("Failed to emulate code using unicorn engine")
                log.error("Code")
                log.error(SourceLine.write_multiline(code))
                raise SelfTestException(
                    f"Selftest failed: Unicorn failed to emulate code: {str(e)}"
                ) from e

            final_register_contents = {}
            for r in regs:
                ur = config.arch.RegisterType.unicorn_reg_by_name(r)
                if ur is None:
                    continue
                final_register_contents[r] = mu.reg_read(ur)
            final_memory_contents = mu.mem_read(RAM_BASE, RAM_SZ)

            return final_register_contents, final_memory_contents

        def failure_dump():
            log.error("Selftest failed")
            log.error("Input code:")
            log.error(SourceLine.write_multiline(codeA))
            log.error("Output code:")
            log.error(SourceLine.write_multiline(codeB))
            log.error("Output registers:")
            log.error(output_registers)

        for _ in range(iterations):
            initial_memory = os.urandom(RAM_SZ)
            initial_stack = os.urandom(STACK_SZ)
            cur_ram = RAM_BASE
            # Set initial register contents arbitrarily, except for registers
            # which must hold valid memory addresses.
            initial_register_contents = {}
            for r in regs:
                initial_register_contents[r] = int.from_bytes(
                    os.urandom(16), byteorder="little"
                )
            for reg, sz in address_registers.items():
                # allocate 2*sz and place pointer in the middle
                # this makes sure that memory can be accessed at negative offsets
                initial_register_contents[reg] = cur_ram + sz
                cur_ram += 2 * sz

            final_regs_old, final_mem_old = run_code(codeA, txt="old")
            final_regs_new, final_mem_new = run_code(codeB, txt="new")

            # Check if memory contents are the same
            if final_mem_old != final_mem_new:
                failure_dump()
                raise SelfTestException("Selftest failed: Memory mismatch")

            # Check that callee-saved registers are the same
            for r in output_registers:
                # skip over hints
                if r.startswith("hint_"):
                    continue
                if final_regs_old[r] != final_regs_new[r]:
                    failure_dump()
                    raise SelfTestException(
                        f"Selftest failed: Register mismatch for {r}: "
                        f"{hex(final_regs_old[r])} != {hex(final_regs_new[r])}"
                    )

        if fnsym is None:
            log.info("Local selftest: OK")
        else:
            log.info(f"Global selftest for {fnsym}: OK")


class Permutation:
    """Helper class for manipulating permutations"""

    @staticmethod
    def is_permutation(perm, sz):
        """Checks whether dictionary perm is a permutation of size sz."""
        err = False
        k = list(perm.keys())
        k.sort()
        v = list(perm.values())
        v.sort()
        if k != list(range(sz)):
            err = True
        if v != list(range(sz)):
            err = True
        if err:
            print(f"Keys:   {k}")
            print(f"Values: {v}")
        return err is False

    @staticmethod
    def permutation_id(sz):
        """Return the identity permutation of size sz."""
        return {i: i for i in range(sz)}

    @staticmethod
    def permutation_comp(p_b, p_a):
        """Compose two permutations.

        This computes 'p_b o p_a', that is, 'p_a first, then p_b'."""
        l_a = len(p_a.values())
        l_b = len(p_b.values())
        assert l_a == l_b
        return {i: p_b[p_a[i]] for i in range(l_a)}

    @staticmethod
    def permutation_pad(perm, pre, post):
        """Pad permutation with identity permutation at front and back"""
        s = len(perm.values())
        r = {}
        r = r | {pre + i: pre + j for (i, j) in perm.items() if isinstance(i, int)}
        r = r | {i: i for i in range(pre)}
        r = r | {i: i for i in map(lambda i: i + s + pre, range(post))}
        return r

    @staticmethod
    def permutation_move_entry_forward(ll, idx_from, idx_to):
        """Create transposition permutation"""
        assert idx_to <= idx_from
        res = {}
        res = res | {i: i for i in range(idx_to)}
        res = res | {i: i + 1 for i in range(idx_to, idx_from)}
        res = res | {idx_from: idx_to}
        res = res | {i: i for i in range(idx_from + 1, ll)}
        return res

    @staticmethod
    def iter_swaps(p, n):
        """Iterate over all inputs that have their order reversed by
        the permutation."""
        return (
            (i, j, p[i], p[j])
            for i in range(n)
            for j in range(n)
            if i < j and p[j] < p[i]
        )


class DeferHandler(logging.Handler):
    """Handler collecting all records produced by a logger and relaying
    them to the same or different logger later."""

    def __init__(self):
        super().__init__()
        self._records = []

    def emit(self, record):
        self._records.append(record)

    def forward(self, logger):
        """Send all captured records to the given logger."""
        for r in self._records:
            logger.handle(r)

    def forward_to_file(self, log_label, filename, lvl=logging.DEBUG):
        """Store all captured records in a file."""
        logger = logging.getLogger(log_label)
        logger.setLevel(lvl)
        h = logging.FileHandler(filename)
        h.setLevel(lvl)
        logger.addHandler(h)
        self.forward(logger)


class Loop(ABC):
    def __init__(self, lbl_start="1", lbl_end="2", loop_init="lr"):
        self.lbl_start = lbl_start
        self.lbl_end = lbl_end
        self.loop_init = loop_init
        self.additional_data = {}

    @abstractmethod
    def start(self, loop_cnt, indentation=0, fixup=0, unroll=1, jump_if_empty=None):
        """Emit starting instruction(s) and jump label for loop"""
        pass

    @abstractmethod
    def end(self, other, indentation=0):
        """Emit compare-and-branch at the end of the loop"""
        pass

    def _extract(self, source, lbl):
        """Locate a loop with start label `lbl` in `source`.```"""
        assert isinstance(source, list)

        # additional_data will be assigned according to the capture groups from
        # loop_end_regexp.
        pre = []
        body = []
        post = []
        # candidate lines for the end of the loop
        loop_end_candidates = []
        loop_lbl_regexp_txt = self.lbl_regex
        loop_lbl_regexp = re.compile(loop_lbl_regexp_txt)

        # end_regex shall contain group cnt as the counter variable.
        # Transform all loop_end_regexp into a list of tuples, where the second
        # element determines whether the instruction should be counted into the
        # body or not. The default is to not put the loop end into the body (False).
        loop_end_regexp_txt = [
            e if isinstance(e, tuple) else (e, False) for e in self.end_regex
        ]
        loop_end_regexp = [re.compile(txt[0]) for txt in loop_end_regexp_txt]
        lines = iter(source)
        line = None
        keep = False
        state = 0  # 0: haven't found loop yet, 1: extracting loop, 2: after loop
        loop_end_ctr = 0
        while True:
            if not keep:
                line = next(lines, None)
            keep = False
            if line is None:
                break
            l_str = line.text
            assert isinstance(line, str) is False
            if state == 0:
                p = loop_lbl_regexp.match(l_str)
                if p is not None and p.group("label") == lbl:
                    line = line.copy().set_text(p.group("remainder"))
                    keep = True
                    state = 1
                else:
                    pre.append(line)
                continue
            if state == 1:
                p = loop_end_regexp[loop_end_ctr].match(l_str)
                if p is not None:
                    # Case: We may have encountered part of the loop end
                    # collect all named groups
                    self.additional_data = self.additional_data | p.groupdict()
                    if loop_end_regexp_txt[loop_end_ctr][1]:
                        # Put all instructions into the loop body, there won't be a
                        # boundary.
                        body.append(line)
                    else:
                        loop_end_candidates.append(line)
                    loop_end_ctr += 1
                    if loop_end_ctr == len(loop_end_regexp):
                        state = 2
                    continue
                elif loop_end_ctr > 0 and l_str != "":
                    # Case: The sequence of loop end candidates was interrupted
                    #       i.e., we found a false-positive or this is not a proper loop

                    # The loop end candidates are not part of the loop, meaning
                    # they belonged to the body
                    # Retain the current line as a candidate that broke the
                    # sequence as it may be the start of the loop end itself
                    keep = True
                    body += loop_end_candidates
                    self.additional_data = {}
                    loop_end_ctr = 0
                    loop_end_candidates = []
                if not keep:
                    body.append(line)
                continue
            if state == 2:
                loop_end_candidates = []
                post.append(line)
                continue
        if state < 2:
            raise FatalParsingException(f"Couldn't identify loop {lbl}")
        return pre, body, post, lbl, self.additional_data

    @staticmethod
    def extract(arch: any, source: list, lbl: str, forced_loop_type: any = None) -> any:
        """
        Find a loop with start label `lbl` in `source` and return it together
        with its type.

        :param arch: module representing architecture for which loops are extracted
        :type arch: any
        :param source: list of SourceLine objects
        :type source: list
        :param lbl: label of the loop to extract
        :type lbl: str
        :param forced_loop_type: if not None, only try to extract this type of loop
        :type forced_loop_type: any
        :return: The extracted loop.
        :rtype: any
        :raises FatalParsingException: If loop with label lbl cannot be found.
        """
        if forced_loop_type is not None:
            loop_types = [forced_loop_type]
        else:
            loop_types = Loop.__subclasses__()
        for loop_type in loop_types:
            # Skip loop type that does not match the architecture
            if not arch.__name__ == loop_type.__module__:
                continue
            try:
                lt = loop_type(lbl)
                # concatenate the extracted loop with an instance of the
                # identified loop_type, (l,) creates a tuple with one element to
                # merge with the tuple retuned by _extract
                return lt._extract(source, lbl) + (lt,)
            except FatalParsingException:
                logging.debug("Parsing loop type '%s'failed", loop_type)
                pass

        raise FatalParsingException(f"Couldn't identify loop {lbl}")


def lookup_multidict(d: dict, inst: any, instclass: any) -> any:
    """
    Lookup a value in a multidict based on an instruction instance.

    :param d: The multidict mapping keys (classes, callables, or tuples) to values.
    :type d: dict
    :param inst: The instruction instance to match against the keys.
    :type inst: any
    :param instclass: A descriptive name of the instruction class, used in error messages.
    :type instclass: any
    :return: The value corresponding to the only matching key in the multidict.
    :rtype: any
    :raises UnknownInstruction: If no matching key found, or multiple matches found.
    """

    result = None
    for ll, v in d.items():
        # Multidict entries can be the following:
        # - An instruction class. It matches any instruction of that class.
        # - A callable. It matches any instruction returning `True` when passed
        #   to the callable.
        # - A tuple of instruction classes or callables. It matches any instruction
        #   which matches at least one element in the tuple.
        def match(x):
            if inspect.isclass(x):
                return isinstance(inst, x)
            assert callable(x)
            return x(inst)

        if not isinstance(ll, tuple):
            ll = [ll]
        for lp in ll:
            if match(lp):
                if result is not None:
                    raise UnknownInstruction(
                        f"Multiple matches found for {instclass} for {inst}"
                    )
                result = v
                break

    if result is None:
        raise UnknownInstruction(f"Couldn't find {instclass} for {inst}")
    return result



#
# Copyright (c) 2022 Arm Limited
# Copyright (c) 2022 Hanno Becker
# Copyright (c) 2023 Amin Abdulrahman, Matthias Kannwischer
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Hanno Becker <hannobecker@posteo.de>
#

"""SLOTHY optimizer

SLOTHY - Super Lazy Optimization of Tricky Handwritten assemblY - is a
fixed-instruction assembly superoptimizer based on constraint solving.
It takes handwritten assembly as input and simultaneously super-optimizes:

* Instruction scheduling
* Register allocation
* Software pipelining

SLOTHY enables a development workflow where developers write 'clean' assembly by hand,
emphasizing the logic of the computation, while SLOTHY automates
microarchitecture-specific micro-optimizations. Since SLOTHY does not change instructions,
and scheduling/allocation optimizations are tightly controlled through configurable and
extensible constraints, the developer keeps close control over the final assembly, while
being freed from the most tedious and readability- and verifiability-impeding
micro-optimizations.

This module provides the Slothy class, which is a stateful interface to both
one-shot and heuristic optimizations using SLOTHY.
"""

import logging
import sys
from types import SimpleNamespace

from slothy.core.dataflow import DataFlowGraph as DFG
from slothy.core.dataflow import Config as DFGConfig, ComputationNode
from slothy.core.core import Config
from slothy.core.heuristics import Heuristics
from slothy.helper import (
    AsmAllocation,
    AsmMacro,
    AsmHelper,
    AsmIfElse,
    CPreprocessor,
    SourceLine,
    LLVM_Mca,
    LLVM_Mca_Error,
    SelfTest,
)


class Slothy:
    """SLOTHY optimizer

    This class provides a stateful interface to both one-shot and heuristic
    optimizations using SLOTHY.

    The basic flow of operation is the following:

    * Initialize an instance, providing models to the target architecture
      and microarchitecture as arguments.
    * Load source code from file or raw string.
    * Repeat: Adjust configuration and conduct an optimization of a loop body or
      straightline block of code, using optimize() or optimize_loop().
    * Write source code to file or raw string.

    The use of heuristics is controlled through the configuration.
    """

    # Quick convenience access to architecture and target from the config
    def _get_arch(self):
        return self.config.arch

    def _get_target(self):
        return self.config.target

    arch = property(_get_arch)
    target = property(_get_target)

    def __init__(self, arch, target, logger=None):
        # Configure default logging to stdout if no handlers are configured
        if logger is None:
            root_logger = logging.getLogger()
            if not root_logger.handlers:
                logging.basicConfig(stream=sys.stdout, level=logging.INFO)
            self.logger = logging.getLogger("slothy")
        else:
            self.logger = logger

        self.config = Config(arch, target, logger=logger)

        # The source, once loaded, is represented as a list of strings
        self._source = None
        self._original_source = None
        self.results = None

        self.last_result = None
        self.success = None

    def _get_version(self):
        try:
            from importlib.metadata import version

            return version("slothy")
        except Exception:
            return "unknown"

    @property
    def source(self):
        """Returns the current source code as an array of SourceLine objects.

        If you want the current source code as a multiline string, use
        get_source_as_string().
        """
        return self._source

    @property
    def original_source(self):
        """Returns the original source code as an array of SourceLine objects

        If you want the current source code as a multiline string, use
        get_original_source_as_string().
        """
        return self._original_source

    @source.setter
    def source(self, val):
        assert SourceLine.is_source(val)
        self._source = val

    @original_source.setter
    def original_source(self, val):
        assert SourceLine.is_source(val)
        self._original_source = val

    def get_source_as_string(self, comments=True, indentation=True, tags=True):
        """Retrieve current source code as multi-line string"""
        return SourceLine.write_multiline(
            self.source, comments=comments, indentation=indentation, tags=tags
        )

    def get_original_source_as_string(self, comments=True, indentation=True, tags=True):
        """Retrieve original source code as multi-line string"""
        return SourceLine.write_multiline(
            self.original_source, comments=comments, indentation=indentation, tags=tags
        )

    def set_source_as_string(self, s):
        """Provide input source code as multi-line string"""
        assert isinstance(s, str)
        reduce = not self.config.ignore_tags
        self.source = SourceLine.read_multiline(s, reduce=reduce)
        if self.original_source is None:
            self.original_source = self.source

    def load_source_raw(self, source):
        """Load source code from multi-line string"""
        self.set_source_as_string(source)
        self.results = []

    def load_source_from_file(self, filename):
        """Load source code from file"""
        if self.source is not None:
            self.logger.warning("Overwriting previous source code")
        with open(filename, "r", encoding="utf8") as f:
            self.load_source_raw(f.read())

    def write_source_to_file(self, filename):
        """Write current source code to file"""
        with open(filename, "w", encoding="utf8") as f:
            f.write(self.get_source_as_string())

    def rename_function(self, old_funcname, new_funcname):
        """Rename a function in the current source code"""
        self.source = AsmHelper.rename_function(self.source, old_funcname, new_funcname)
        self.source = AsmHelper.rename_function(
            self.source, "_" + old_funcname, "_" + new_funcname
        )

    @staticmethod
    def _dump(name, s, logger, err=False):
        assert isinstance(s, list)
        fun = logger.debug if not err else logger.error
        fun(f"Dump: {name}")
        for line in s:
            fun(f"> {line.to_string()}")

    def global_selftest(
        self, funcname: str, address_registers: any, iterations: int = 5
    ) -> None:
        """Conduct a function-level selftest
            .. important::

            To run this, you need `llvm-nm`, `llvm-readobj`, `llvm-mc`
            in your PATH. Those are part of a standard LLVM setup.

        :param funcname: Name of function to be called. Must be exposed as a symbol
        :type funcname: str
        :param address_registers: Dictionary indicating which GPRs are pointers to
            buffers of which size.
            For example, `{ "x0": 1024, "x4": 1024 }` would indicate that both x0 and x4
            point to buffers of size 1024 bytes. The global selftest needs to know this to
            setup valid calls to the assembly routine.
        :type address_registers: any
        :param iterations: Number of iterations to run the selftest for.
        :type iterations: int
        :return: Returns early of target architecture of not supported.
        :rtype: None
        """

        log = self.logger.getChild(f"global_selftest_{funcname}")

        if (
            self.config.arch.unicorn_arch is None
            or self.config.arch.llvm_mc_arch is None
        ):
            log.warning("Selftest not supported on target architecture")
            return

        old_source = self.original_source
        new_source = self.source

        SelfTest.run(
            self.config,
            log,
            old_source,
            new_source,
            address_registers,
            self.config.arch.RegisterType.callee_saved_registers(),
            iterations,
            fnsym=funcname,
        )

    #
    # Stateful wrappers around heuristics
    #

    def unfold(self, start=None, end=None, macros=True, aliases=False):
        """Unfold macros and/or register aliases in specified region"""
        pre, body, post = AsmHelper.extract(self.source, start, end)

        aliases = AsmAllocation.parse_allocs(pre)
        c = self.config.copy()
        c.add_aliases(aliases)

        if c.with_preprocessor:
            self.logger.info("Apply C preprocessor...")
            body = CPreprocessor.unfold(
                pre, body, post, c.compiler_binary, include=c.compiler_include_paths
            )
            self.logger.debug("Code after preprocessor:")
            Slothy._dump("preprocessed", body, self.logger, err=False)

        body = SourceLine.split_semicolons(body)

        # Unfold macros
        if macros is True:
            body = AsmMacro.unfold_all_macros(
                pre, body, inherit_comments=c.inherit_macro_comments
            )

        # Unfold register aliases
        if aliases is True:
            body = AsmAllocation.unfold_all_aliases(c.register_aliases, body)

        # Add labels again
        if start is not None:
            body = [SourceLine(f"{start}:")] + body
        if end is not None:
            body = body + [SourceLine(f"{end}:")]

        self.source = pre + body + post

    def _make_llvm_mca_stats(self, pre, code, post, txt, indentation):
        try:
            code = CPreprocessor.unfold(
                pre,
                code,
                post,
                self.config.compiler_binary,
                include=self.config.compiler_include_paths,
            )
            if self.config.llvm_mca_issue_width_overwrite is True:
                issue_width = self.config.target.issue_rate
            else:
                issue_width = None
            stats = LLVM_Mca.run(
                pre,
                code,
                self.config.arch.llvm_mca_arch,
                self.config.target.llvm_mca_target,
                self.logger,
                full=self.config.llvm_mca_full,
                issue_width=issue_width,
            )
            stats = (
                ["", f"LLVM MCA STATISTICS ({txt}) BEGIN", ""]
                + stats
                + ["", f"ORIGINAL LLVM MCA STATISTICS ({txt}) END", ""]
            )
            stats = [SourceLine("").add_comment(r) for r in stats]
            stats = SourceLine.apply_indentation(stats, indentation)
        except LLVM_Mca_Error:
            self.logger.warning("Failed to run LLVM-MCA -- ignoring")
            stats = []
        return stats

    def optimize(
        self,
        start: str = None,
        end: str = None,
        loop_synthesis_cb: any = None,
        logname: str = None,
    ):
        """Optimize all or part of the currently loaded source code

        .. note::

            It is OK to use this in software pipelining mode. In this case, the
            tool will output preamble, kernel, and postamble separately, while the looping
            code itself needs to be introduced by the user. Alternatively, a callback can
            be provided which will be given preamble, kernel, postamble, and the number
            of exceptional iterations, and piece together a list of source code lines
            from that.


        :param start: The label marking the beginning of the part of the code to optimize.
              This cannot be used together with the 'loop' argument.
        :type start: str
        :param end: The label marking the end of the part of the code to optimize.
              This cannot be used together with the 'loop' argument.
        :type end: str
        :param loop_synthesis_cb: Optional (None by default) callback synthesis final
               source code from tuple of (preamble, kernel, postamble, # exceptional
               iterations).
        :type loop_synthesis_cb: any
        :param logname: Optional name of the logger.
        :type logname: str
        """
        if logname is None and start is not None:
            logname = start
        if logname is None and end is not None:
            logname = end

        logger = self.logger.getChild(logname) if logname is not None else self.logger
        pre, body, post = AsmHelper.extract(self.source, start, end)

        aliases = AsmAllocation.parse_allocs(pre)
        c = self.config.copy()
        c.add_aliases(aliases)

        # Check if the body has a dominant indentation
        indentation = AsmHelper.find_indentation(body)

        if c.with_preprocessor:
            self.logger.info("Apply C preprocessor...")
            body = CPreprocessor.unfold(
                pre, body, post, c.compiler_binary, include=c.compiler_include_paths
            )
            self.logger.debug("Code after preprocessor:")
            Slothy._dump("preprocessed", body, self.logger, err=False)

        body = SourceLine.split_semicolons(body)
        body = AsmMacro.unfold_all_macros(
            pre, body, inherit_comments=c.inherit_macro_comments
        )
        body = AsmAllocation.unfold_all_aliases(c.register_aliases, body)
        body = AsmIfElse.process_instructions(body)
        body = SourceLine.apply_indentation(body, indentation)
        logger.info("SLOTHY version: %s", self._get_version())
        self.logger.info("Instructions in body: %d", SourceLine.instruction_count(body))

        if self.config.with_llvm_mca_before is True:
            orig_stats = self._make_llvm_mca_stats(
                pre, body, post, "ORIGINAL", indentation
            )

        early, core, late, num_exceptional = Heuristics.periodic(body, logger, c)

        if self.config.with_llvm_mca_before is True:
            core = core + orig_stats

        if self.config.with_llvm_mca_after is True:
            new_stats_kernel = self._make_llvm_mca_stats(
                pre, core, post, "OPTIMIZED", indentation
            )

            core = core + new_stats_kernel

        def indented(code):
            return [SourceLine(line).set_indentation(indentation) for line in code]

        if start is not None:
            core = [SourceLine(f"{start}:")] + core
        if end is not None:
            core += [SourceLine(f"{end}:")]

        core = SourceLine.apply_indentation(core, self.config.indentation)
        if not self.config.sw_pipelining.enabled:
            assert early == []
            assert late == []
            assert num_exceptional == 0
            optimized_source = core
        elif loop_synthesis_cb is not None:
            optimized_source = loop_synthesis_cb(pre, core, post, num_exceptional)
        else:
            optimized_source = []
            optimized_source += indented(
                [f"// Exceptional iterations: {num_exceptional}", "// Preamble"]
            )
            optimized_source += early
            optimized_source += indented(["// Kernel"])
            optimized_source += core
            optimized_source += indented(["// Postamble"])
            optimized_source += late

        self.source = pre + optimized_source + post
        assert SourceLine.is_source(self.source)

    def get_loop_input_output(
        self, loop_lbl: str, forced_loop_type: any = None
    ) -> list:
        """Find all registers that a loop body depends on

        :param loop_lbl: Label of loop to process.
        :type loop_lbl: str
        :param forced_loop_type: Forces the loop to be parsed as a certain type.
        :type forced_loop_type: any
        :return: list all inputs of the loop.
        :rtype: list
        """
        logger = self.logger.getChild(loop_lbl)
        _, body, _, _, _ = self.arch.Loop.extract(
            self.arch, self.source, loop_lbl, forced_loop_type=forced_loop_type
        )

        c = self.config.copy()
        dfgc = DFGConfig(c)
        dfgc.inputs_are_outputs = True
        return list(DFG(body, logger.getChild("dfg_kernel_deps"), dfgc).inputs)

    def get_input_from_output(self, start, end, outputs=None):
        """For the piece of straightline code, infer which input registers affect its
        output"""
        if outputs is None:
            outputs = {}
        logger = self.logger.getChild(f"{start}_{end}_infer_input")
        pre, body, _ = AsmHelper.extract(self.source, start, end)

        aliases = AsmAllocation.parse_allocs(pre)
        c = self.config.copy()
        c.add_aliases(aliases)
        c.outputs = outputs

        body = AsmMacro.unfold_all_macros(
            pre, body, inherit_comments=c.inherit_macro_comments
        )
        body = AsmAllocation.unfold_all_aliases(c.register_aliases, body)
        dfgc = DFGConfig(c)
        return list(DFG(body, logger.getChild("dfg_find_deps"), dfgc).inputs)

    def _fusion_core(self, pre, body, post, logger, ssa=True):
        c = self.config.copy()

        if c.with_preprocessor:
            self.logger.info("Apply C preprocessor...")
            body = CPreprocessor.unfold(
                pre, body, post, c.compiler_binary, include=c.compiler_include_paths
            )
            self.logger.debug("Code after preprocessor:")
            Slothy._dump("preprocessed", body, self.logger, err=False)
        body = SourceLine.split_semicolons(body)

        aliases = AsmAllocation.parse_allocs(pre)
        c.add_aliases(aliases)

        body = AsmMacro.unfold_all_macros(
            pre, body, inherit_comments=c.inherit_macro_comments
        )
        body = AsmAllocation.unfold_all_aliases(c.register_aliases, body)
        dfgc = DFGConfig(c)

        if ssa is True:
            dfg = DFG(body, logger.getChild("ssa"), dfgc, parsing_cb=False)
            dfg.ssa()
            body = [ComputationNode.to_source_line(t) for t in dfg.nodes]

        dfg = DFG(body, logger.getChild("fusion"), dfgc, parsing_cb=False)
        dfg.apply_fusion_cbs()
        body = [ComputationNode.to_source_line(t) for t in dfg.nodes]

        return body

    def fusion_region(self, start: str, end: str, **kwargs: any):
        """Run fusion callbacks on straightline code replacing certain
        instruction (sequences) with an alternative. These replacements are
        defined in the architectural model by setting an instruction class'
        global_fusion_cb.

        :param start: The label marking the beginning of the part of the code to
              apply fusion to.
        :type start: str
        :param end: The label marking the end of the part of the code to apply
              fusion to.
        :type end: str
        :param **kwargs: Additional arguments to pass to the fusion callbacks.
        :type **kwargs: any
        """
        logger = self.logger.getChild(f"ssa_{start}_{end}")
        pre, body, post = AsmHelper.extract(self.source, start, end)

        body_ssa = (
            [SourceLine(f"{start}:")]
            + self._fusion_core(pre, body, post, logger, **kwargs)
            + [SourceLine(f"{end}:")]
        )
        self.source = pre + body_ssa + post
        assert SourceLine.is_source(self.source)

    def fusion_loop(self, loop_lbl: str, forced_loop_type: any = None, **kwargs: any):
        """Run fusion callbacks on loop body replacing certain instruction
        (sequences) with an alternative. These replacements are defined in the
        architectural model by setting an instruction class' global_fusion_cb.

        :param loop_lbl: Label of loop to which the fusions are applied to.
        :type loop_lbl: str
        :param forced_loop_type: Forces the loop to be parsed as a certain type.
        :type forced_loop_type: any
        :param **kwargs: Additional arguments to pass to the fusion callbacks.
        :type **kwargs: any
        """
        logger = self.logger.getChild(f"ssa_loop_{loop_lbl}")

        pre, body, post, _, other_data, loop = self.arch.Loop.extract(
            self.arch, self.source, loop_lbl, forced_loop_type=forced_loop_type
        )

        try:
            loop_cnt = other_data["cnt"]
        except KeyError:
            loop_cnt = None

        indentation = AsmHelper.find_indentation(body)

        body_ssa = (
            SourceLine.read_multiline(loop.start(loop_cnt))
            + SourceLine.apply_indentation(
                self._fusion_core(pre, body, post, logger, **kwargs), indentation
            )
            + SourceLine.read_multiline(loop.end(other_data))
        )

        self.source = pre + body_ssa + post
        assert SourceLine.is_source(self.source)

    def optimize_loop(
        self, loop_lbl: str, postamble_label: str = None, forced_loop_type: any = None
    ):
        """Optimize the loop starting at a given label

        :param loop_lbl: Label of loop to optimize.
        :type loop_lbl: str
        :param postamble_label: Marks end of loop kernel.
        :type postamble_label: str
        :param forced_loop_type: Forces the loop to be parsed as a certain type.
        :type forced_loop_type: any
        """

        logger = self.logger.getChild(loop_lbl)

        early, body, late, _, other_data, loop = self.arch.Loop.extract(
            self.arch, self.source, loop_lbl, forced_loop_type=forced_loop_type
        )

        try:
            loop_cnt = other_data["cnt"]
        except KeyError:
            loop_cnt = None

        # Check if the body has a dominant indentation
        indentation = AsmHelper.find_indentation(body)

        aliases = AsmAllocation.parse_allocs(early)
        c = self.config.copy()
        c.add_aliases(aliases)

        if c.with_preprocessor:
            self.logger.info("Apply C preprocessor...")
            body = CPreprocessor.unfold(
                early, body, late, c.compiler_binary, include=c.compiler_include_paths
            )
            self.logger.debug("Code after preprocessor:")
            Slothy._dump("preprocessed", body, self.logger, err=False)

        body = SourceLine.split_semicolons(body)
        body = AsmMacro.unfold_all_macros(
            early, body, inherit_comments=c.inherit_macro_comments
        )
        body = AsmAllocation.unfold_all_aliases(c.register_aliases, body)
        body = AsmIfElse.process_instructions(body)
        body = SourceLine.apply_indentation(body, indentation)
        logger.info("SLOTHY version: %s", self._get_version())
        self.logger.info(
            "Optimizing loop %s (%d instructions) ...",
            loop_lbl,
            SourceLine.instruction_count(body),
        )

        if self.config.with_llvm_mca_before is True:
            orig_stats = self._make_llvm_mca_stats(
                early, body, late, "ORIGINAL", indentation
            )

        preamble_code, kernel_code, postamble_code, num_exceptional = (
            Heuristics.periodic(body, logger, c)
        )

        # Remove branch instructions from preamble and postamble
        postamble_code = [
            line for line in postamble_code if not line.tags.get("branch")
        ]
        postamble_code = [
            line for line in postamble_code if not line.tags.get("branch")
        ]

        if self.config.with_llvm_mca_before is True:
            kernel_code = kernel_code + orig_stats

        if self.config.with_llvm_mca_after is True:
            new_stats_kernel = self._make_llvm_mca_stats(
                early, kernel_code, late, "OPTIMIZED", indentation
            )
            kernel_code = kernel_code + new_stats_kernel

            if (
                self.config.sw_pipelining.optimize_preamble is True
                and len(preamble_code) > 0
            ):
                new_stats_preamble = self._make_llvm_mca_stats(
                    early, preamble_code, late, "PREAMBLE", indentation
                )
                preamble_code = preamble_code + new_stats_preamble

            if (
                self.config.sw_pipelining.optimize_postamble is True
                and len(postamble_code) > 0
            ):
                new_stats_postamble = self._make_llvm_mca_stats(
                    early, postamble_code, late, "POSTAMBLE", indentation
                )
                postamble_code = postamble_code + new_stats_postamble

        def indented(code):
            if not SourceLine.is_source(code):
                code = SourceLine.read_multiline(code)
            return SourceLine.apply_indentation(code, self.config.indentation)

        loop_lbl_end = f"{loop_lbl}_end"

        def loop_lbl_iter(i):
            return f"{loop_lbl}_iter_{i}"

        optimized_code = []

        if self.config.sw_pipelining.unknown_iteration_count:
            for i in range(1, num_exceptional):
                optimized_code += indented(
                    self.arch.Branch.if_equal(loop_cnt, i, loop_lbl_iter(i))
                )

        optimized_code += indented(preamble_code)

        if self.config.sw_pipelining.unknown_iteration_count:
            if postamble_label is None:
                postamble_label = f"{loop_lbl}_postamble"
            jump_if_empty = postamble_label
        else:
            jump_if_empty = None

        optimized_code += SourceLine.read_multiline(
            loop.start(
                loop_cnt,
                indentation=self.config.indentation,
                fixup=num_exceptional,
                unroll=self.config.sw_pipelining.unroll,
                jump_if_empty=jump_if_empty,
                preamble_code=preamble_code,
                body_code=kernel_code,
                postamble_code=postamble_code,
                register_aliases=c.register_aliases,
            )
        )
        optimized_code += indented(kernel_code)
        optimized_code += SourceLine.read_multiline(
            loop.end(other_data, indentation=self.config.indentation)
        )
        if postamble_label is not None:
            optimized_code += [
                SourceLine(f"{postamble_label}:").add_comment("end of loop kernel")
            ]
        optimized_code += indented(postamble_code)

        if self.config.sw_pipelining.unknown_iteration_count:
            optimized_code += indented(self.arch.Branch.unconditional(loop_lbl_end))
            for i in range(1, num_exceptional):
                exceptional = i * body
                c2 = c.copy()
                c2.sw_pipelining.enabled = False
                res = Heuristics.linear(
                    exceptional, logger.getChild(f"exceptional_{i}"), c2
                )
                optimized_code += [SourceLine(f"{loop_lbl_iter(i)}:")]
                optimized_code += indented(res.code)
                optimized_code += [SourceLine(f"{loop_lbl_iter(i)}_end:")]
                if i != num_exceptional - 1:
                    optimized_code += indented(
                        self.arch.Branch.unconditional(loop_lbl_end)
                    )
            optimized_code += [SourceLine(f"{loop_lbl_end}:")]

        self.last_result = SimpleNamespace()
        dfgc = DFGConfig(c)
        dfgc.inputs_are_outputs = True
        self.last_result.kernel_input_output = list(
            DFG(kernel_code, logger.getChild("dfg_kernel_deps"), dfgc).inputs
        )

        self.source = early + optimized_code + late
        self.success = True



#
# Copyright (c) 2022 Arm Limited
# Copyright (c) 2022 Hanno Becker
# Copyright (c) 2023 Amin Abdulrahman, Matthias Kannwischer
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Hanno Becker <hannobecker@posteo.de>
#

"""
Experimental and highly incomplete model capturing an approximation of the
frontend limitations and latencies of the Neoverse N1 CPU
"""

from enum import Enum
from slothy.helper import lookup_multidict
from slothy.targets.aarch64.aarch64_neon import (
    is_neon_instruction,
    find_class,
    all_subclass_leaves,
    Ldp_X,
    Ldr_X,
    Str_X,
    Stp_X,
    Ldr_D,
    Ldr_Q,
    Str_Q,
    Stp_Q,
    Ldp_Q,
    Vrev,
    uaddlp,
    vmov,
    vmovi,
    vadd,
    vxtn,
    vusra,
    vmul,
    vdup,
    AESInstruction,
    Transpose,
    AArch64NeonLogical,
    VShiftImmediateBasic,
    AArch64BasicArithmetic,
    AArch64ConditionalSelect,
    AArch64ConditionalCompare,
    AArch64Logical,
    AArch64LogicalShifted,
    AArch64Move,
    AArch64Shift,
    Tst,
    AArch64ShiftedArithmetic,
    AArch64HighMultiply,
    AArch64Multiply,
    VecToGprMov,
    St3,
    St4,
    Vzip,
    vsub,
    vsrshr,
    Vmul,
    Vmla,
    Vqdmulh,
    Vmull,
    Vmlal,
    umull_wform,
    vmul_lane,
    vmla,
    vmla_lane,
    AArch64NeonCount,
    vmls,
    ASimdCompare,
    vmls_lane,
    vext,
    AArch64NeonShiftInsert,
    vtbl,
    vuaddlv_sform,
    fmov_s_form,  # from vec to gen reg
    fmov_0,
    fmov_0_force_output,
    fmov_1,
    fmov_1_force_output,
    q_ldr1_stack,
    Q_Ld2_Lane_Post_Inc,
    St2,
)

issue_rate = 4
llvm_mca_target = "neoverse-n1"


class ExecutionUnit(Enum):
    """Enumeration of execution units in approximative Neoverse-N1 SLOTHY model"""

    SCALAR_I0 = 0
    SCALAR_I1 = 1
    SCALAR_I2 = 2
    SCALAR_M = 2  # Overlaps with third I pipeline
    LSU0 = 3
    LSU1 = 4
    VEC0 = 5
    VEC1 = 6

    def __repr__(self):
        return self.name

    @classmethod
    def I(cls):  # noqa: E743
        return [
            ExecutionUnit.SCALAR_I0,
            ExecutionUnit.SCALAR_I1,
            ExecutionUnit.SCALAR_I2,
        ]

    @classmethod
    def M(cls):
        return [ExecutionUnit.SCALAR_M]

    @classmethod
    def V(cls):
        return [ExecutionUnit.VEC0, ExecutionUnit.VEC1]

    @classmethod
    def V0(cls):
        return [ExecutionUnit.VEC0]

    @classmethod
    def V1(cls):
        return [ExecutionUnit.VEC1]

    @classmethod
    def LSU(cls):
        return [ExecutionUnit.LSU0, ExecutionUnit.LSU1]


# Opaque functions called by SLOTHY to add further microarchitecture-
# specific constraints which are not encapsulated by the general framework.
def add_further_constraints(slothy):
    if slothy.config.constraints.functional_only:
        return
    slothy.restrict_slots_for_instructions_by_property(is_neon_instruction, [0, 1])
    slothy.restrict_slots_for_instructions_by_property(
        lambda t: is_neon_instruction(t) is False, [1, 2, 3]
    )


def has_min_max_objective(config):
    _ = config
    return False


def get_min_max_objective(slothy):
    _ = slothy


execution_units = {
    (
        Ldp_X,
        Ldr_X,
        Str_X,
        Stp_X,
        Ldr_D,
        Ldr_Q,
        Str_Q,
        Stp_Q,
        Ldp_Q,
    ): ExecutionUnit.LSU(),
    # TODO: The following would be more accurate, but does not
    #       necessarily lead to better results, while making the
    #       optimization slower. Investigate...
    #
    # Ldr_Q)            : ExecutionUnit.LSU(),
    # Str_Q : [[ExecutionUnit.VEC0, ExecutionUnit.LSU0],
    #          [ExecutionUnit.VEC0, ExecutionUnit.LSU1],
    #          [ExecutionUnit.VEC1, ExecutionUnit.LSU0],
    #          [ExecutionUnit.VEC1, ExecutionUnit.LSU1]],
    # TODO: As above, this should somehow occupy both V and L
    St2: ExecutionUnit.V(),
    St3: ExecutionUnit.V(),
    St4: ExecutionUnit.V(),
    ASimdCompare: ExecutionUnit.V(),
    vtbl: ExecutionUnit.V(),
    (Vzip, Vrev, uaddlp): ExecutionUnit.V(),
    AArch64NeonCount: ExecutionUnit.V(),
    (vmov): ExecutionUnit.V(),
    VecToGprMov: ExecutionUnit.V(),
    Transpose: ExecutionUnit.V(),
    (vmovi): ExecutionUnit.V(),
    (vadd, vsub): ExecutionUnit.V(),
    (vxtn): ExecutionUnit.V(),
    VShiftImmediateBasic: ExecutionUnit.V1(),
    (
        AArch64NeonShiftInsert,
        vsrshr,
    ): ExecutionUnit.V1(),
    vusra: ExecutionUnit.V1(),
    AESInstruction: ExecutionUnit.V0(),
    (Vmul, Vmla, Vqdmulh, Vmull, Vmlal): ExecutionUnit.V0(),
    AArch64NeonLogical: ExecutionUnit.V(),
    vext: ExecutionUnit.V(),
    (
        AArch64BasicArithmetic,
        AArch64ConditionalSelect,
        AArch64ConditionalCompare,
        AArch64Logical,
        AArch64LogicalShifted,
        AArch64Move,
    ): ExecutionUnit.I(),
    AArch64Shift: ExecutionUnit.I(),
    Tst: ExecutionUnit.I(),
    AArch64ShiftedArithmetic: ExecutionUnit.M(),
    (
        fmov_0,
        fmov_0_force_output,
        fmov_1,
        fmov_1_force_output,
    ): ExecutionUnit.M(),
    fmov_s_form: ExecutionUnit.V1(),  # from vec to gen reg
    umull_wform: ExecutionUnit.M(),
    (AArch64HighMultiply, AArch64Multiply): ExecutionUnit.M(),
    vdup: ExecutionUnit.M(),
    # 8B/8H occupies both V0, V1
    vuaddlv_sform: [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    q_ldr1_stack: ExecutionUnit.V(),
    Q_Ld2_Lane_Post_Inc: ExecutionUnit.V(),
}

inverse_throughput = {
    (Ldr_X, Str_X, Ldr_D, Ldr_Q, Str_Q, Ldp_Q): 1,
    (Ldp_X, Stp_X): 2,
    AArch64NeonCount: 1,
    Stp_Q: 2,
    St3: 3,  # Multiple structures, Q form, storing bytes
    St4: 6,  # TODO: Really??
    (Vzip, uaddlp, Vrev): 1,
    VecToGprMov: 1,
    (vadd, vsub): 1,
    (vmov): 1,
    ASimdCompare: 1,
    Transpose: 1,
    AESInstruction: 1,
    AArch64NeonLogical: 1,
    vext: 1,
    (vmovi): 1,
    (vxtn): 1,
    VShiftImmediateBasic: 1,
    (AArch64NeonShiftInsert, vsrshr): 1,
    (Vmul, Vmla, Vqdmulh): 2,
    vusra: 1,
    vtbl: 1,
    (Vmull, Vmlal): 1,
    (
        AArch64BasicArithmetic,
        AArch64ConditionalSelect,
        AArch64ConditionalCompare,
        AArch64Logical,
        AArch64LogicalShifted,
        AArch64Move,
    ): 1,
    AArch64Shift: 1,
    AArch64ShiftedArithmetic: 1,
    Tst: 1,
    (
        fmov_0,
        fmov_0_force_output,
        fmov_1,
        fmov_1_force_output,
    ): 1,
    fmov_s_form: 1,  # from vec to gen reg
    (AArch64HighMultiply): 4,
    (AArch64Multiply): 3,
    (vdup): 1,
    umull_wform: 1,
    vuaddlv_sform: 1,  # 8B/8H
    q_ldr1_stack: 1,
    Q_Ld2_Lane_Post_Inc: 2,
    St2: 2,
}

default_latencies = {
    # For OOO uArch we use relaxed latency modeling for load instructions
    # since the uArch will heavily front-load them anyway
    (Ldp_X, Ldr_X, Ldr_D, Ldr_Q, Stp_Q, Ldp_Q): 4,
    (Stp_X, Str_X, Str_Q): 2,
    St3: 6,  # Multiple structures, Q form, storing bytes
    St4: 4,
    (Vzip, Vrev, uaddlp): 2,
    VecToGprMov: 2,
    ASimdCompare: 2,
    (vxtn): 2,
    AArch64NeonCount: 2,
    AESInstruction: 2,
    AArch64NeonLogical: 2,
    vext: 2,
    Transpose: 2,
    (vadd, vsub): 2,
    (vmov): 2,  # ???
    (vmovi): 2,
    (Vmul, Vmla, Vqdmulh): 5,
    vusra: 4,  # TODO: Add fwd path
    (Vmull, Vmlal): 4,
    VShiftImmediateBasic: 2,
    AArch64NeonShiftInsert: 2,
    (vsrshr): 4,
    (
        AArch64BasicArithmetic,
        AArch64ConditionalSelect,
        AArch64ConditionalCompare,
        AArch64Logical,
        AArch64LogicalShifted,
        AArch64Move,
    ): 1,
    AArch64Shift: 1,
    AArch64ShiftedArithmetic: 2,
    Tst: 1,
    (
        fmov_0,
        fmov_0_force_output,
        fmov_1,
        fmov_1_force_output,
    ): 3,
    fmov_s_form: 2,  # from vec to gen reg
    AArch64HighMultiply: 5,
    AArch64Multiply: 4,
    (vdup): 3,
    umull_wform: 2,
    vtbl: 2,
    vuaddlv_sform: 5,  # 8B/8H
    q_ldr1_stack: 7,
    Q_Ld2_Lane_Post_Inc: 7,
    St2: 4,
}


def get_latency(src, out_idx, dst):
    _ = out_idx  # out_idx unused

    instclass_src = find_class(src)
    instclass_dst = find_class(dst)

    latency = lookup_multidict(default_latencies, src, instclass_src)

    # Fast mul->mla forwarding
    if (
        instclass_src in [vmul, vmul_lane]
        and instclass_dst in [vmla, vmla_lane, vmls, vmls_lane]
        and src.args_out[0] == dst.args_in_out[0]
    ):
        return 2
    # Fast mla->mla forwarding
    if (
        instclass_src in [vmla, vmla_lane, vmls, vmls_lane]
        and instclass_dst in [vmla, vmla_lane, vmls, vmls_lane]
        and src.args_in_out[0] == dst.args_in_out[0]
    ):
        return 2
    # Fast mull->mlal forwarding
    if (
        instclass_src in all_subclass_leaves(Vmull)
        and instclass_dst in all_subclass_leaves(Vmlal)
        and src.args_out[0] == dst.args_in_out[0]
    ):
        return 1
    # Fast mlal->mlal forwarding
    if (
        instclass_src in all_subclass_leaves(Vmlal)
        and instclass_dst in all_subclass_leaves(Vmlal)
        and src.args_in_out[0] == dst.args_in_out[0]
    ):
        return 1

    return latency


def get_units(src):
    instclass_src = find_class(src)
    units = lookup_multidict(execution_units, src, instclass_src)
    if isinstance(units, list):
        return units
    return [units]


def get_inverse_throughput(src):
    instclass_src = find_class(src)
    return lookup_multidict(inverse_throughput, src, instclass_src)
